#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
import unittest
from datetime import datetime
from unittest.mock import MagicMock, Mock

import pytest
from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, AirbyteStream, Level, SyncMode
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.file_based.availability_strategy import DefaultFileBasedAvailabilityStrategy
from airbyte_cdk.sources.file_based.config.csv_format import CsvFormat
from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.discovery_policy import DefaultDiscoveryPolicy
from airbyte_cdk.sources.file_based.exceptions import FileBasedErrorsCollector
from airbyte_cdk.sources.file_based.file_types import default_parsers
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.schema_validation_policies import EmitRecordPolicy
from airbyte_cdk.sources.file_based.stream import DefaultFileBasedStream
from airbyte_cdk.sources.file_based.stream.concurrent.adapters import (
    FileBasedStreamFacade,
    FileBasedStreamPartition,
    FileBasedStreamPartitionGenerator,
)
from airbyte_cdk.sources.file_based.stream.concurrent.cursor import FileBasedNoopCursor
from airbyte_cdk.sources.message import InMemoryMessageRepository
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor
from airbyte_cdk.sources.streams.concurrent.exceptions import ExceptionWithDisplayMessage
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_cdk.sources.utils.slice_logger import SliceLogger
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer
from freezegun import freeze_time

_ANY_SYNC_MODE = SyncMode.full_refresh
_ANY_STATE = {"state_key": "state_value"}
_ANY_CURSOR_FIELD = ["a", "cursor", "key"]
_STREAM_NAME = "stream"
_ANY_CURSOR = Mock(spec=FileBasedNoopCursor)


@pytest.mark.parametrize(
    "sync_mode",
    [
        pytest.param(SyncMode.full_refresh, id="test_full_refresh"),
        pytest.param(SyncMode.incremental, id="test_incremental"),
    ],
)
def test_file_based_stream_partition_generator(sync_mode):
    stream = Mock()
    message_repository = Mock()
    stream_slices = [{"files": [RemoteFile(uri="1", last_modified=datetime.now())]},
                     {"files": [RemoteFile(uri="2", last_modified=datetime.now())]}]
    stream.stream_slices.return_value = stream_slices

    partition_generator = FileBasedStreamPartitionGenerator(stream, message_repository, _ANY_SYNC_MODE, _ANY_CURSOR_FIELD, _ANY_STATE, _ANY_CURSOR)

    partitions = list(partition_generator.generate())
    slices = [partition.to_slice() for partition in partitions]
    assert slices == stream_slices
    stream.stream_slices.assert_called_once_with(sync_mode=_ANY_SYNC_MODE, cursor_field=_ANY_CURSOR_FIELD, stream_state=_ANY_STATE)


@pytest.mark.parametrize(
    "transformer, expected_records",
    [
        pytest.param(
            TypeTransformer(TransformConfig.NoTransform),
            [Record({"data": "1"}, _STREAM_NAME), Record({"data": "2"}, _STREAM_NAME)],
            id="test_no_transform",
        ),
        pytest.param(
            TypeTransformer(TransformConfig.DefaultSchemaNormalization),
            [Record({"data": 1}, _STREAM_NAME), Record({"data": 2}, _STREAM_NAME)],
            id="test_default_transform",
        ),
    ],
)
def test_file_based_stream_partition(transformer, expected_records):
    stream = Mock()
    stream.name = _STREAM_NAME
    stream.get_json_schema.return_value = {"type": "object", "properties": {"data": {"type": ["integer"]}}}
    stream.transformer = transformer
    message_repository = InMemoryMessageRepository()
    _slice = None
    sync_mode = SyncMode.full_refresh
    cursor_field = None
    state = None
    partition = FileBasedStreamPartition(stream, _slice, message_repository, sync_mode, cursor_field, state, _ANY_CURSOR)

    a_log_message = AirbyteMessage(
        type=MessageType.LOG,
        log=AirbyteLogMessage(
            level=Level.INFO,
            message='slice:{"partition": 1}',
        ),
    )

    stream_data = [a_log_message, {"data": "1"}, {"data": "2"}]
    stream.read_records.return_value = stream_data

    records = list(partition.read())
    messages = list(message_repository.consume_queue())

    assert records == expected_records
    assert messages == [a_log_message]


@pytest.mark.parametrize(
    "exception_type, expected_display_message",
    [
        pytest.param(Exception, None, id="test_exception_no_display_message"),
        pytest.param(ExceptionWithDisplayMessage, "display_message", id="test_exception_no_display_message"),
    ],
)
def test_file_based_stream_partition_raising_exception(exception_type, expected_display_message):
    stream = Mock()
    stream.get_error_display_message.return_value = expected_display_message

    message_repository = InMemoryMessageRepository()
    _slice = None

    partition = FileBasedStreamPartition(stream, _slice, message_repository, _ANY_SYNC_MODE, _ANY_CURSOR_FIELD, _ANY_STATE, _ANY_CURSOR)

    stream.read_records.side_effect = Exception()

    with pytest.raises(exception_type) as e:
        list(partition.read())
        if isinstance(e, ExceptionWithDisplayMessage):
            assert e.display_message == "display message"


@freeze_time("2023-06-09T00:00:00Z")
@pytest.mark.parametrize(
    "_slice, expected_hash",
    [
        pytest.param({"files": [RemoteFile(uri="1", last_modified=datetime.strptime("2023-06-09T00:00:00Z", "%Y-%m-%dT%H:%M:%SZ"))]}, hash(("stream", "2023-06-09T00:00:00.000000Z_1")), id="test_hash_with_slice"),
        pytest.param(None, hash("stream"), id="test_hash_no_slice"),
    ],
)
def test_file_based_stream_partition_hash(_slice, expected_hash):
    stream = Mock()
    stream.name = "stream"
    partition = FileBasedStreamPartition(stream, _slice, Mock(), _ANY_SYNC_MODE, _ANY_CURSOR_FIELD, _ANY_STATE, _ANY_CURSOR)

    _hash = partition.__hash__()
    assert _hash == expected_hash


class StreamFacadeTest(unittest.TestCase):
    def setUp(self):
        self._abstract_stream = Mock()
        self._abstract_stream.name = "stream"
        self._abstract_stream.as_airbyte_stream.return_value = AirbyteStream(
            name="stream",
            json_schema={"type": "object"},
            supported_sync_modes=[SyncMode.full_refresh],
        )
        self._legacy_stream = DefaultFileBasedStream(
            cursor=FileBasedNoopCursor(MagicMock()),
            config=FileBasedStreamConfig(name="stream", format=CsvFormat()),
            catalog_schema={},
            stream_reader=MagicMock(),
            availability_strategy=DefaultFileBasedAvailabilityStrategy(MagicMock()),
            discovery_policy=DefaultDiscoveryPolicy(),
            parsers=default_parsers,
            validation_policy=EmitRecordPolicy(),
            errors_collector=FileBasedErrorsCollector(),
        )
        self._cursor = Mock(spec=Cursor)
        self._logger = Mock()
        self._slice_logger = Mock()
        self._slice_logger.should_log_slice_message.return_value = False
        self._facade = FileBasedStreamFacade(self._abstract_stream, self._legacy_stream, self._cursor, self._slice_logger, self._logger)
        self._source = Mock()

        self._stream = Mock()
        self._stream.primary_key = "id"

    def test_name_is_delegated_to_wrapped_stream(self):
        assert self._facade.name == self._abstract_stream.name

    def test_cursor_field_is_a_string(self):
        self._abstract_stream.cursor_field = "cursor_field"
        assert self._facade.cursor_field == "cursor_field"

    def test_source_defined_cursor_is_true(self):
        assert self._facade.source_defined_cursor

    def test_json_schema_is_delegated_to_wrapped_stream(self):
        json_schema = {"type": "object"}
        self._abstract_stream.get_json_schema.return_value = json_schema
        assert self._facade.get_json_schema() == json_schema
        self._abstract_stream.get_json_schema.assert_called_once_with()

    def test_given_cursor_is_noop_when_supports_incremental_then_return_legacy_stream_response(self):
        assert (
            FileBasedStreamFacade(
                self._abstract_stream, self._legacy_stream, _ANY_CURSOR, Mock(spec=SliceLogger), Mock(spec=logging.Logger)
            ).supports_incremental
            == self._legacy_stream.supports_incremental
        )

    def test_given_cursor_is_not_noop_when_supports_incremental_then_return_true(self):
        assert FileBasedStreamFacade(
            self._abstract_stream, self._legacy_stream, Mock(spec=Cursor), Mock(spec=SliceLogger), Mock(spec=logging.Logger)
        ).supports_incremental

    def test_full_refresh(self):
        expected_stream_data = [{"data": 1}, {"data": 2}]
        records = [Record(data, "stream") for data in expected_stream_data]

        partition = Mock()
        partition.read.return_value = records
        self._abstract_stream.generate_partitions.return_value = [partition]

        actual_stream_data = list(self._facade.read_records(SyncMode.full_refresh, None, {}, None))

        assert actual_stream_data == expected_stream_data

    def test_read_records_full_refresh(self):
        expected_stream_data = [{"data": 1}, {"data": 2}]
        records = [Record(data, "stream") for data in expected_stream_data]
        partition = Mock()
        partition.read.return_value = records
        self._abstract_stream.generate_partitions.return_value = [partition]

        actual_stream_data = list(self._facade.read_full_refresh(None, None, None))

        assert actual_stream_data == expected_stream_data

    def test_read_records_incremental(self):
        expected_stream_data = [{"data": 1}, {"data": 2}]
        records = [Record(data, "stream") for data in expected_stream_data]
        partition = Mock()
        partition.read.return_value = records
        self._abstract_stream.generate_partitions.return_value = [partition]

        actual_stream_data = list(self._facade.read_incremental(None, None, None, None, None, None, None))

        assert actual_stream_data == expected_stream_data

    def test_create_from_stream_stream(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = "id"
        stream.cursor_field = "cursor"

        facade = FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)

        assert facade.name == "stream"
        assert facade.cursor_field == "cursor"
        assert facade._abstract_stream._primary_key == ["id"]

    def test_create_from_stream_stream_with_none_primary_key(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = None
        stream.cursor_field = []

        facade = FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)
        assert facade._abstract_stream._primary_key == []

    def test_create_from_stream_with_composite_primary_key(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = ["id", "name"]
        stream.cursor_field = []

        facade = FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)
        assert facade._abstract_stream._primary_key == ["id", "name"]

    def test_create_from_stream_with_empty_list_cursor(self):
        stream = Mock()
        stream.primary_key = "id"
        stream.cursor_field = []

        facade = FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)

        assert facade.cursor_field == []

    def test_create_from_stream_raises_exception_if_primary_key_is_nested(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = [["field", "id"]]

        with self.assertRaises(ValueError):
            FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)

    def test_create_from_stream_raises_exception_if_primary_key_has_invalid_type(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = 123

        with self.assertRaises(ValueError):
            FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)

    def test_create_from_stream_raises_exception_if_cursor_field_is_nested(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = "id"
        stream.cursor_field = ["field", "cursor"]

        with self.assertRaises(ValueError):
            FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)

    def test_create_from_stream_with_cursor_field_as_list(self):
        stream = Mock()
        stream.name = "stream"
        stream.primary_key = "id"
        stream.cursor_field = ["cursor"]

        facade = FileBasedStreamFacade.create_from_stream(stream, self._source, self._logger, _ANY_STATE, self._cursor)
        assert facade.cursor_field == "cursor"

    def test_create_from_stream_none_message_repository(self):
        self._stream.name = "stream"
        self._stream.primary_key = "id"
        self._stream.cursor_field = "cursor"
        self._source.message_repository = None

        with self.assertRaises(ValueError):
            FileBasedStreamFacade.create_from_stream(self._stream, self._source, self._logger, {}, self._cursor)

    def test_get_error_display_message_no_display_message(self):
        self._stream.get_error_display_message.return_value = "display_message"

        facade = FileBasedStreamFacade.create_from_stream(self._stream, self._source, self._logger, _ANY_STATE, self._cursor)

        expected_display_message = None
        e = Exception()

        display_message = facade.get_error_display_message(e)

        assert expected_display_message == display_message

    def test_get_error_display_message_with_display_message(self):
        self._stream.get_error_display_message.return_value = "display_message"

        facade = FileBasedStreamFacade.create_from_stream(self._stream, self._source, self._logger, _ANY_STATE, self._cursor)

        expected_display_message = "display_message"
        e = ExceptionWithDisplayMessage("display_message")

        display_message = facade.get_error_display_message(e)

        assert expected_display_message == display_message


@pytest.mark.parametrize(
    "exception, expected_display_message",
    [
        pytest.param(Exception("message"), None, id="test_no_display_message"),
        pytest.param(ExceptionWithDisplayMessage("message"), "message", id="test_no_display_message"),
    ],
)
def test_get_error_display_message(exception, expected_display_message):
    stream = Mock()
    legacy_stream = Mock()
    cursor = Mock(spec=Cursor)
    facade = FileBasedStreamFacade(stream, legacy_stream, cursor, Mock().Mock(), Mock())

    display_message = facade.get_error_display_message(exception)

    assert display_message == expected_display_message



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.declarative.extractors.record_filter import RecordFilter


@pytest.mark.parametrize(
    "test_name, filter_template, records, expected_records",
    [
        (
            "test_using_state_filter",
            "{{ record['created_at'] > stream_state['created_at'] }}",
            [{"id": 1, "created_at": "06-06-21"}, {"id": 2, "created_at": "06-07-21"}, {"id": 3, "created_at": "06-08-21"}],
            [{"id": 2, "created_at": "06-07-21"}, {"id": 3, "created_at": "06-08-21"}],
        ),
        (
            "test_with_slice_filter",
            "{{ record['last_seen'] >= stream_slice['last_seen'] }}",
            [{"id": 1, "last_seen": "06-06-21"}, {"id": 2, "last_seen": "06-07-21"}, {"id": 3, "last_seen": "06-10-21"}],
            [{"id": 3, "last_seen": "06-10-21"}],
        ),
        (
            "test_with_next_page_token_filter",
            "{{ record['id'] >= next_page_token['last_seen_id'] }}",
            [{"id": 11}, {"id": 12}, {"id": 13}, {"id": 14}, {"id": 15}],
            [{"id": 14}, {"id": 15}],
        ),
        (
            "test_missing_filter_fields_return_no_results",
            "{{ record['id'] >= next_page_token['path_to_nowhere'] }}",
            [{"id": 11}, {"id": 12}, {"id": 13}, {"id": 14}, {"id": 15}],
            [],
        ),
        (
            "test_using_parameters_filter",
            "{{ record['created_at'] > parameters['created_at'] }}",
            [{"id": 1, "created_at": "06-06-21"}, {"id": 2, "created_at": "06-07-21"}, {"id": 3, "created_at": "06-08-21"}],
            [{"id": 3, "created_at": "06-08-21"}],
        ),
    ],
)
def test_record_filter(test_name, filter_template, records, expected_records):
    config = {"response_override": "stop_if_you_see_me"}
    parameters = {"created_at": "06-07-21"}
    stream_state = {"created_at": "06-06-21"}
    stream_slice = {"last_seen": "06-10-21"}
    next_page_token = {"last_seen_id": 14}
    record_filter = RecordFilter(config=config, condition=filter_template, parameters=parameters)

    actual_records = record_filter.filter_records(
        records, stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token
    )
    assert actual_records == expected_records



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.declarative.auth.selective_authenticator import SelectiveAuthenticator


def test_authenticator_selected(mocker):
    authenticators = {"one": mocker.Mock(), "two": mocker.Mock()}
    auth = SelectiveAuthenticator(
        config={"auth": {"type": "one"}},
        authenticators=authenticators,
        authenticator_selection_path=["auth", "type"],
    )

    assert auth is authenticators["one"]


def test_selection_path_not_found(mocker):
    authenticators = {"one": mocker.Mock(), "two": mocker.Mock()}

    with pytest.raises(ValueError, match="The path from `authenticator_selection_path` is not found in the config"):
        _ = SelectiveAuthenticator(
            config={"auth": {"type": "one"}},
            authenticators=authenticators,
            authenticator_selection_path=["auth_type"],
        )


def test_selected_auth_not_found(mocker):
    authenticators = {"one": mocker.Mock(), "two": mocker.Mock()}

    with pytest.raises(ValueError, match="The authenticator `unknown` is not found"):
        _ = SelectiveAuthenticator(
            config={"auth": {"type": "unknown"}},
            authenticators=authenticators,
            authenticator_selection_path=["auth", "type"],
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import datetime
import decimal

import pyarrow as pa
from airbyte_cdk.utils.traced_exception import AirbyteTracedException
from unit_tests.sources.file_based.in_memory_files_source import TemporaryParquetFilesStreamReader
from unit_tests.sources.file_based.scenarios.file_based_source_builder import FileBasedSourceBuilder
from unit_tests.sources.file_based.scenarios.scenario_builder import TestScenarioBuilder

_single_parquet_file = {
    "a.parquet": {
        "contents": [
            ("col1", "col2"),
            ("val11", "val12"),
            ("val21", "val22"),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    }
}

_single_partitioned_parquet_file = {
    "path_prefix/partition1=1/partition2=2/a.parquet": {
        "contents": [
            ("col1", "col2"),
            ("val11", "val12"),
            ("val21", "val22"),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    }
}

_parquet_file_with_decimal = {
    "a.parquet": {
        "contents": [
            ("col1",),
            (decimal.Decimal("13.00"),),
        ],
        "schema": pa.schema(
            [
                pa.field("col1", pa.decimal128(5, 2)),
            ]
        ),
        "last_modified": "2023-06-05T03:54:07.000Z",
    }
}

_multiple_parquet_file = {
    "a.parquet": {
        "contents": [
            ("col1", "col2"),
            ("val11a", "val12a"),
            ("val21a", "val22a"),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    },
    "b.parquet": {
        "contents": [
            ("col1", "col2", "col3"),
            ("val11b", "val12b", "val13b"),
            ("val21b", "val22b", "val23b"),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    },
}

_parquet_file_with_various_types = {
    "a.parquet": {
        "contents": [
            (
                "col_bool",
                "col_int8",
                "col_int16",
                "col_int32",
                "col_uint8",
                "col_uint16",
                "col_uint32",
                "col_uint64",
                "col_float32",
                "col_float64",
                "col_string",
                "col_date32",
                "col_date64",
                "col_timestamp_without_tz",
                "col_timestamp_with_tz",
                "col_time32s",
                "col_time32ms",
                "col_time64us",
                "col_struct",
                "col_list",
                "col_duration",
                "col_binary",
            ),
            (
                True,
                -1,
                1,
                2,
                2,
                3,
                4,
                5,
                3.14,
                5.0,
                "2020-01-01",
                datetime.date(2021, 1, 1),
                datetime.date(2022, 1, 1),
                datetime.datetime(2023, 1, 1, 1, 2, 3),
                datetime.datetime(2024, 3, 4, 5, 6, 7, tzinfo=datetime.timezone.utc),
                datetime.time(1, 2, 3),
                datetime.time(2, 3, 4),
                datetime.time(1, 2, 3, 4),
                {"struct_key": "struct_value"},
                [1, 2, 3, 4],
                12345,
                b"binary string. Hello world!",
            ),
        ],
        "schema": pa.schema(
            [
                pa.field("col_bool", pa.bool_()),
                pa.field("col_int8", pa.int8()),
                pa.field("col_int16", pa.int16()),
                pa.field("col_int32", pa.int32()),
                pa.field("col_uint8", pa.uint8()),
                pa.field("col_uint16", pa.uint16()),
                pa.field("col_uint32", pa.uint32()),
                pa.field("col_uint64", pa.uint64()),
                pa.field("col_float32", pa.float32()),
                pa.field("col_float64", pa.float64()),
                pa.field("col_string", pa.string()),
                pa.field("col_date32", pa.date32()),
                pa.field("col_date64", pa.date64()),
                pa.field("col_timestamp_without_tz", pa.timestamp("s")),
                pa.field("col_timestamp_with_tz", pa.timestamp("s", tz="UTC")),
                pa.field("col_time32s", pa.time32("s")),
                pa.field("col_time32ms", pa.time32("ms")),
                pa.field("col_time64us", pa.time64("us")),
                pa.field("col_struct", pa.struct([pa.field("struct_key", pa.string())])),
                pa.field("col_list", pa.list_(pa.int32())),
                pa.field("col_duration", pa.duration("s")),
                pa.field("col_binary", pa.binary()),
            ]
        ),
        "last_modified": "2023-06-05T03:54:07.000Z",
    }
}

single_parquet_scenario = (
    TestScenarioBuilder()
    .set_name("single_parquet_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "parquet"},
                    "globs": ["*"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_single_parquet_file, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": "val12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": "val22",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "col2": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

single_partitioned_parquet_scenario = (
    TestScenarioBuilder()
    .set_name("single_partitioned_parquet_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "parquet"},
                    "globs": ["path_prefix/**/*"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_single_partitioned_parquet_file, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": "val12",
                    "partition1": "1",
                    "partition2": "2",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "path_prefix/partition1=1/partition2=2/a.parquet",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": "val22",
                    "partition1": "1",
                    "partition2": "2",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "path_prefix/partition1=1/partition2=2/a.parquet",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "col2": {"type": ["null", "string"]},
                            "partition1": {"type": ["null", "string"]},
                            "partition2": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

multi_parquet_scenario = (
    TestScenarioBuilder()
    .set_name("multi_parquet_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "parquet"},
                    "globs": ["*"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_file_type("parquet")
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_multiple_parquet_file, file_type="parquet"))
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "col2": {"type": ["null", "string"]},
                            "col3": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.parquet",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.parquet",
                },
                "stream": "stream1",
            },
        ]
    )
).build()

parquet_various_types_scenario = (
    TestScenarioBuilder()
    .set_name("parquet_various_types")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "parquet"},
                    "globs": ["*"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_parquet_file_with_various_types, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col_bool": {
                                "type": ["null", "boolean"],
                            },
                            "col_int8": {
                                "type": ["null", "integer"],
                            },
                            "col_int16": {
                                "type": ["null", "integer"],
                            },
                            "col_int32": {
                                "type": ["null", "integer"],
                            },
                            "col_uint8": {
                                "type": ["null", "integer"],
                            },
                            "col_uint16": {
                                "type": ["null", "integer"],
                            },
                            "col_uint32": {
                                "type": ["null", "integer"],
                            },
                            "col_uint64": {
                                "type": ["null", "integer"],
                            },
                            "col_float32": {
                                "type": ["null", "number"],
                            },
                            "col_float64": {
                                "type": ["null", "number"],
                            },
                            "col_string": {
                                "type": ["null", "string"],
                            },
                            "col_date32": {"type": ["null", "string"], "format": "date"},
                            "col_date64": {"type": ["null", "string"], "format": "date"},
                            "col_timestamp_without_tz": {"type": ["null", "string"], "format": "date-time"},
                            "col_timestamp_with_tz": {"type": ["null", "string"], "format": "date-time"},
                            "col_time32s": {
                                "type": ["null", "string"],
                            },
                            "col_time32ms": {
                                "type": ["null", "string"],
                            },
                            "col_time64us": {
                                "type": ["null", "string"],
                            },
                            "col_struct": {
                                "type": ["null", "object"],
                            },
                            "col_list": {
                                "type": ["null", "array"],
                            },
                            "col_duration": {
                                "type": ["null", "integer"],
                            },
                            "col_binary": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string",
                            },
                            "_ab_source_file_url": {
                                "type": "string",
                            },
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col_bool": True,
                    "col_int8": -1,
                    "col_int16": 1,
                    "col_int32": 2,
                    "col_uint8": 2,
                    "col_uint16": 3,
                    "col_uint32": 4,
                    "col_uint64": 5,
                    "col_float32": 3.14,
                    "col_float64": 5.0,
                    "col_string": "2020-01-01",
                    "col_date32": "2021-01-01",
                    "col_date64": "2022-01-01",
                    "col_timestamp_without_tz": "2023-01-01T01:02:03",
                    "col_timestamp_with_tz": "2024-03-04T05:06:07+00:00",
                    "col_time32s": "01:02:03",
                    "col_time32ms": "02:03:04",
                    "col_time64us": "01:02:03.000004",
                    "col_struct": {"struct_key": "struct_value"},
                    "col_list": [1, 2, 3, 4],
                    "col_duration": 12345,
                    "col_binary": "binary string. Hello world!",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
        ]
    )
).build()

parquet_file_with_decimal_no_config_scenario = (
    TestScenarioBuilder()
    .set_name("parquet_file_with_decimal_no_config")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "parquet"},
                    "globs": ["*"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_parquet_file_with_decimal, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "13.00",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

parquet_file_with_decimal_as_string_scenario = (
    TestScenarioBuilder()
    .set_name("parquet_file_with_decimal_as_string")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "globs": ["*"],
                    "validation_policy": "Emit Record",
                    "format": {"filetype": "parquet", "decimal_as_float": False},
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_parquet_file_with_decimal, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "13.00",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.parquet",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

parquet_file_with_decimal_as_float_scenario = (
    TestScenarioBuilder()
    .set_name("parquet_file_with_decimal_as_float")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "globs": ["*"],
                    "validation_policy": "Emit Record",
                    "format": {"filetype": "parquet", "decimal_as_float": True},
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_parquet_file_with_decimal, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records(
        [
            {
                "data": {"col1": 13.00, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a.parquet"},
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "number"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

parquet_file_with_decimal_legacy_config_scenario = (
    TestScenarioBuilder()
    .set_name("parquet_file_with_decimal_legacy_config")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {
                        "filetype": "parquet",
                    },
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_parquet_file_with_decimal, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records(
        [
            {
                "data": {"col1": 13.00, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a.parquet"},
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "number"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

parquet_with_invalid_config_scenario = (
    TestScenarioBuilder()
    .set_name("parquet_with_invalid_config")
    .set_config({"streams": [{"name": "stream1", "globs": ["*"], "validation_policy": "Emit Record", "format": {"filetype": "csv"}}]})
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_stream_reader(TemporaryParquetFilesStreamReader(files=_single_parquet_file, file_type="parquet"))
        .set_file_type("parquet")
    )
    .set_expected_records([])
    .set_expected_logs({"read": [{"level": "ERROR", "message": "Error parsing record"}]})
    .set_expected_discover_error(AirbyteTracedException, "Error inferring schema from files")
    .set_expected_read_error(AirbyteTracedException, "Please check the logged errors for more information.")
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "col2": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.streams.http.rate_limiting import default_backoff_handler
from requests import exceptions


def helper_with_exceptions(exception_type):
    raise exception_type


@pytest.mark.parametrize(
    "max_tries, max_time, factor, exception_to_raise",
    [
        (1, None, 1, exceptions.ConnectTimeout),
        (1, 1, 0, exceptions.ReadTimeout),
        (2, 2, 1, exceptions.ConnectionError),
        (3, 3, 1, exceptions.ChunkedEncodingError),
    ],
)
def test_default_backoff_handler(max_tries: int, max_time: int, factor: int, exception_to_raise: Exception):
    backoff_handler = default_backoff_handler(max_tries=max_tries, max_time=max_time, factor=factor)(helper_with_exceptions)
    with pytest.raises(exception_to_raise):
        backoff_handler(exception_to_raise)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.declarative.parsers.manifest_component_transformer import ManifestComponentTransformer


@pytest.mark.parametrize(
    "component, expected_component",
    [
        pytest.param(
            {"type": "DeclarativeSource", "streams": [{"type": "DeclarativeStream", "retriever": {}, "schema_loader": {}}]},
            {
                "type": "DeclarativeSource",
                "streams": [
                    {
                        "type": "DeclarativeStream",
                        "retriever": {"type": "SimpleRetriever"},
                        "schema_loader": {"type": "JsonFileSchemaLoader"},
                    }
                ],
            },
            id="test_declarative_stream",
        ),
        pytest.param(
            {
                "type": "DeclarativeStream",
                "retriever": {"type": "SimpleRetriever", "paginator": {}, "record_selector": {}, "requester": {}},
            },
            {
                "type": "DeclarativeStream",
                "retriever": {
                    "type": "SimpleRetriever",
                    "paginator": {"type": "NoPagination"},
                    "record_selector": {"type": "RecordSelector"},
                    "requester": {"type": "HttpRequester"},
                },
            },
            id="test_simple_retriever",
        ),
        pytest.param(
            {"type": "DeclarativeStream", "requester": {"type": "HttpRequester", "error_handler": {}}},
            {
                "type": "DeclarativeStream",
                "requester": {
                    "type": "HttpRequester",
                    "error_handler": {"type": "DefaultErrorHandler"},
                },
            },
            id="test_http_requester",
        ),
        pytest.param(
            {"type": "SimpleRetriever", "paginator": {"type": "DefaultPaginator", "page_size_option": {}, "page_token_option": {}}},
            {
                "type": "SimpleRetriever",
                "paginator": {
                    "type": "DefaultPaginator",
                    "page_size_option": {"type": "RequestOption"},
                    "page_token_option": {},
                },
            },
            id="test_default_paginator",
        ),
        pytest.param(
            {"type": "SimpleRetriever", "partition_router": {"type": "SubstreamPartitionRouter", "parent_stream_configs": [{}, {}, {}]}},
            {
                "type": "SimpleRetriever",
                "partition_router": {
                    "type": "SubstreamPartitionRouter",
                    "parent_stream_configs": [
                        {"type": "ParentStreamConfig"},
                        {"type": "ParentStreamConfig"},
                        {"type": "ParentStreamConfig"},
                    ],
                },
            },
            id="test_substream_slicer",
        ),
    ],
)
def test_find_default_types(component, expected_component):
    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component


@pytest.mark.parametrize(
    "component, expected_component",
    [
        pytest.param(
            {
                "type": "SimpleRetriever",
                "requester": {"type": "HttpRequester", "authenticator": {"class_name": "source_greenhouse.components.NewAuthenticator"}},
            },
            {
                "type": "SimpleRetriever",
                "requester": {
                    "type": "HttpRequester",
                    "authenticator": {"type": "CustomAuthenticator", "class_name": "source_greenhouse.components.NewAuthenticator"},
                },
            },
            id="test_custom_authenticator",
        ),
        pytest.param(
            {
                "type": "SimpleRetriever",
                "record_selector": {
                    "type": "RecordSelector",
                    "extractor": {"class_name": "source_greenhouse.components.NewRecordExtractor"},
                },
            },
            {
                "type": "SimpleRetriever",
                "record_selector": {
                    "type": "RecordSelector",
                    "extractor": {"type": "CustomRecordExtractor", "class_name": "source_greenhouse.components.NewRecordExtractor"},
                },
            },
            id="test_custom_extractor",
        ),
    ],
)
def test_transform_custom_components(component, expected_component):
    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component


def test_propagate_parameters_to_all_components():
    component = {
        "type": "DeclarativeSource",
        "streams": [
            {
                "type": "DeclarativeStream",
                "$parameters": {"name": "roasters", "primary_key": "id"},
                "retriever": {
                    "type": "SimpleRetriever",
                    "record_selector": {"type": "RecordSelector", "extractor": {"type": "DpathExtractor", "field_path": []}},
                    "requester": {
                        "type": "HttpRequester",
                        "name": '{{ parameters["name"] }}',
                        "url_base": "https://coffee.example.io/v1/",
                        "http_method": "GET",
                    },
                },
            }
        ],
    }

    expected_component = {
        "type": "DeclarativeSource",
        "streams": [
            {
                "type": "DeclarativeStream",
                "retriever": {
                    "type": "SimpleRetriever",
                    "name": "roasters",
                    "primary_key": "id",
                    "record_selector": {
                        "type": "RecordSelector",
                        "extractor": {
                            "type": "DpathExtractor",
                            "field_path": [],
                            "name": "roasters",
                            "primary_key": "id",
                            "$parameters": {"name": "roasters", "primary_key": "id"},
                        },
                        "name": "roasters",
                        "primary_key": "id",
                        "$parameters": {"name": "roasters", "primary_key": "id"},
                    },
                    "requester": {
                        "type": "HttpRequester",
                        "name": '{{ parameters["name"] }}',
                        "url_base": "https://coffee.example.io/v1/",
                        "http_method": "GET",
                        "primary_key": "id",
                        "$parameters": {"name": "roasters", "primary_key": "id"},
                    },
                    "$parameters": {"name": "roasters", "primary_key": "id"},
                },
                "name": "roasters",
                "primary_key": "id",
                "$parameters": {"name": "roasters", "primary_key": "id"},
            }
        ],
    }

    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component


def test_component_parameters_take_precedence_over_parent_parameters():
    component = {
        "type": "DeclarativeStream",
        "retriever": {
            "type": "SimpleRetriever",
            "requester": {
                "type": "HttpRequester",
                "name": "high_priority",
                "url_base": "https://coffee.example.io/v1/",
                "http_method": "GET",
                "primary_key": "id",
                "$parameters": {
                    "name": "high_priority",
                },
            },
            "$parameters": {
                "name": "low_priority",
            },
        },
    }

    expected_component = {
        "type": "DeclarativeStream",
        "retriever": {
            "type": "SimpleRetriever",
            "name": "low_priority",
            "requester": {
                "type": "HttpRequester",
                "name": "high_priority",
                "url_base": "https://coffee.example.io/v1/",
                "http_method": "GET",
                "primary_key": "id",
                "$parameters": {
                    "name": "high_priority",
                },
            },
            "$parameters": {
                "name": "low_priority",
            },
        },
    }

    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component


def test_do_not_propagate_parameters_that_have_the_same_field_name():
    component = {
        "type": "DeclarativeStream",
        "streams": [
            {
                "type": "DeclarativeStream",
                "$parameters": {
                    "name": "roasters",
                    "primary_key": "id",
                    "schema_loader": {"type": "JsonFileSchemaLoader", "file_path": './source_coffee/schemas/{{ parameters["name"] }}.json'},
                },
            }
        ],
    }

    expected_component = {
        "type": "DeclarativeStream",
        "streams": [
            {
                "type": "DeclarativeStream",
                "name": "roasters",
                "primary_key": "id",
                "schema_loader": {
                    "type": "JsonFileSchemaLoader",
                    "file_path": './source_coffee/schemas/{{ parameters["name"] }}.json',
                    "name": "roasters",
                    "primary_key": "id",
                    "$parameters": {
                        "name": "roasters",
                        "primary_key": "id",
                    },
                },
                "$parameters": {
                    "name": "roasters",
                    "primary_key": "id",
                    "schema_loader": {"type": "JsonFileSchemaLoader", "file_path": './source_coffee/schemas/{{ parameters["name"] }}.json'},
                },
            }
        ],
    }

    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component


def test_ignore_empty_parameters():
    component = {
        "type": "DeclarativeStream",
        "retriever": {
            "type": "SimpleRetriever",
            "record_selector": {"type": "RecordSelector", "extractor": {"type": "DpathExtractor", "field_path": []}},
        },
    }

    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == component


def test_only_propagate_parameters_to_components():
    component = {
        "type": "ParentComponent",
        "component_with_object_properties": {
            "type": "TestComponent",
            "subcomponent": {
                "type": "TestSubComponent",
                "some_field": "high_priority",
                "$parameters": {
                    "some_option": "already",
                },
            },
            "dictionary_field": {"details": "should_not_contain_parameters", "other": "no_parameters_as_fields"},
            "$parameters": {
                "included": "not!",
            },
        },
    }

    expected_component = {
        "type": "ParentComponent",
        "component_with_object_properties": {
            "type": "TestComponent",
            "subcomponent": {
                "type": "TestSubComponent",
                "some_field": "high_priority",
                "some_option": "already",
                "included": "not!",
                "$parameters": {"some_option": "already", "included": "not!"},
            },
            "dictionary_field": {"details": "should_not_contain_parameters", "other": "no_parameters_as_fields"},
            "included": "not!",
            "$parameters": {
                "included": "not!",
            },
        },
    }

    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component


def test_do_not_propagate_parameters_on_json_schema_object():
    component = {
        "type": "DeclarativeStream",
        "streams": [
            {
                "type": "DeclarativeStream",
                "schema_loader": {
                    "type": "InlineSchemaLoader",
                    "schema": {
                        "type": "object",
                        "$schema": "http://json-schema.org/schema#",
                        "properties": {"id": {"type": "string"}},
                    },
                },
                "$parameters": {
                    "name": "roasters",
                    "primary_key": "id",
                },
            }
        ],
    }

    expected_component = {
        "type": "DeclarativeStream",
        "streams": [
            {
                "type": "DeclarativeStream",
                "name": "roasters",
                "primary_key": "id",
                "schema_loader": {
                    "type": "InlineSchemaLoader",
                    "name": "roasters",
                    "primary_key": "id",
                    "schema": {
                        "type": "object",
                        "$schema": "http://json-schema.org/schema#",
                        "properties": {"id": {"type": "string"}},
                    },
                    "$parameters": {
                        "name": "roasters",
                        "primary_key": "id",
                    },
                },
                "$parameters": {
                    "name": "roasters",
                    "primary_key": "id",
                },
            }
        ],
    }

    transformer = ManifestComponentTransformer()
    actual_component = transformer.propagate_types_and_parameters("", component, {})

    assert actual_component == expected_component



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, List, Mapping, Union

import requests


@dataclass
class Decoder:
    """
    Decoder strategy to transform a requests.Response into a Mapping[str, Any]
    """

    @abstractmethod
    def decode(self, response: requests.Response) -> Union[Mapping[str, Any], List]:
        """
        Decodes a requests.Response into a Mapping[str, Any] or an array
        :param response: the response to decode
        :return: Mapping or array describing the response
        """
        pass



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
import json
from copy import deepcopy
from pathlib import Path as FilePath
from typing import Any, Dict, Optional, Union
from unittest import TestCase
from unittest.mock import Mock

import pytest
from airbyte_cdk.test.mock_http.response import HttpResponse
from airbyte_cdk.test.mock_http.response_builder import (
    FieldPath,
    FieldUpdatePaginationStrategy,
    HttpResponseBuilder,
    NestedPath,
    PaginationStrategy,
    Path,
    RecordBuilder,
    create_record_builder,
    create_response_builder,
    find_template,
)

_RECORDS_FIELD = "records_field"
_ID_FIELD = "record_id"
_CURSOR_FIELD = "record_cursor"
_ANY_RECORD = {"a_record_field": "a record value"}
_SOME_RECORDS = {_RECORDS_FIELD: [_ANY_RECORD]}
_A_RESPONSE_TEMPLATE = _SOME_RECORDS

_RECORD_BUILDER = 0
_RESPONSE_BUILDER = 1


def _record_builder(
    response_template: Dict[str, Any],
    records_path: Union[FieldPath, NestedPath],
    record_id_path: Optional[Path] = None,
    record_cursor_path: Optional[Union[FieldPath, NestedPath]] = None,
) -> RecordBuilder:
    return create_record_builder(deepcopy(response_template), records_path, record_id_path, record_cursor_path)


def _any_record_builder() -> RecordBuilder:
    return create_record_builder({"record_path": [{"a_record": "record value"}]}, FieldPath("record_path"))


def _response_builder(
    response_template: Dict[str, Any],
    records_path: Union[FieldPath, NestedPath],
    pagination_strategy: Optional[PaginationStrategy] = None
) -> HttpResponseBuilder:
    return create_response_builder(deepcopy(response_template), records_path, pagination_strategy=pagination_strategy)


def _body(response: HttpResponse) -> Dict[str, Any]:
    return json.loads(response.body)


class RecordBuilderTest(TestCase):
    def test_given_with_id_when_build_then_set_id(self) -> None:
        builder = _record_builder({_RECORDS_FIELD: [{_ID_FIELD: "an id"}]}, FieldPath(_RECORDS_FIELD), FieldPath(_ID_FIELD))
        record = builder.with_id("another id").build()
        assert record[_ID_FIELD] == "another id"

    def test_given_nested_id_when_build_then_set_id(self) -> None:
        builder = _record_builder({_RECORDS_FIELD: [{"nested": {_ID_FIELD: "id"}}]}, FieldPath(_RECORDS_FIELD), NestedPath(["nested", _ID_FIELD]))
        record = builder.with_id("another id").build()
        assert record["nested"][_ID_FIELD] == "another id"

    def test_given_id_path_not_provided_but_with_id_when_build_then_raise_error(self) -> None:
        builder = _record_builder(_A_RESPONSE_TEMPLATE, FieldPath(_RECORDS_FIELD), None)
        with pytest.raises(ValueError):
            builder.with_id("any_id").build()

    def test_given_no_id_in_template_for_path_when_build_then_raise_error(self) -> None:
        with pytest.raises(ValueError):
            _record_builder({_RECORDS_FIELD: [{"record without id": "should fail"}]}, FieldPath(_RECORDS_FIELD), FieldPath(_ID_FIELD))

    def test_given_with_cursor_when_build_then_set_id(self) -> None:
        builder = _record_builder(
            {_RECORDS_FIELD: [{_CURSOR_FIELD: "a cursor"}]},
            FieldPath(_RECORDS_FIELD),
            record_cursor_path=FieldPath(_CURSOR_FIELD)
        )
        record = builder.with_cursor("another cursor").build()
        assert record[_CURSOR_FIELD] == "another cursor"

    def test_given_nested_cursor_when_build_then_set_cursor(self) -> None:
        builder = _record_builder(
            {_RECORDS_FIELD: [{"nested": {_CURSOR_FIELD: "a cursor"}}]},
            FieldPath(_RECORDS_FIELD),
            record_cursor_path=NestedPath(["nested", _CURSOR_FIELD])
        )
        record = builder.with_cursor("another cursor").build()
        assert record["nested"][_CURSOR_FIELD] == "another cursor"

    def test_given_with_field_when_build_then_write_field(self) -> None:
        builder = _any_record_builder()
        record = builder.with_field(FieldPath("to_write_field"), "a field value").build()
        assert record["to_write_field"] == "a field value"

    def test_given_nested_cursor_when_build_then_write_field(self) -> None:
        builder = _any_record_builder()
        record = builder.with_field(NestedPath(["path", "to_write_field"]), "a field value").build()
        assert record["path"]["to_write_field"] == "a field value"

    def test_given_cursor_path_not_provided_but_with_id_when_build_then_raise_error(self) -> None:
        builder = _record_builder(_A_RESPONSE_TEMPLATE, FieldPath(_RECORDS_FIELD))
        with pytest.raises(ValueError):
            builder.with_cursor("any cursor").build()

    def test_given_no_cursor_in_template_for_path_when_build_then_raise_error(self) -> None:
        with pytest.raises(ValueError):
            _record_builder(
                {_RECORDS_FIELD: [{"record without cursor": "should fail"}]},
                FieldPath(_RECORDS_FIELD),
                record_cursor_path=FieldPath(_ID_FIELD)
            )


class HttpResponseBuilderTest(TestCase):
    def test_given_records_in_template_but_no_with_records_when_build_then_no_records(self) -> None:
        builder = _response_builder({_RECORDS_FIELD: [{"a_record_field": "a record value"}]}, FieldPath(_RECORDS_FIELD))
        response = builder.build()
        assert len(_body(response)[_RECORDS_FIELD]) == 0

    def test_given_many_records_when_build_then_response_has_records(self) -> None:
        builder = _response_builder(_A_RESPONSE_TEMPLATE, FieldPath(_RECORDS_FIELD))
        a_record_builder = Mock(spec=RecordBuilder)
        a_record_builder.build.return_value = {"a record": 1}
        another_record_builder = Mock(spec=RecordBuilder)
        another_record_builder.build.return_value = {"another record": 2}

        response = builder.with_record(a_record_builder).with_record(another_record_builder).build()

        assert len(_body(response)[_RECORDS_FIELD]) == 2

    def test_when_build_then_default_status_code_is_200(self) -> None:
        builder = _response_builder(_A_RESPONSE_TEMPLATE, FieldPath(_RECORDS_FIELD))
        response = builder.build()
        assert response.status_code == 200

    def test_given_status_code_when_build_then_status_code_is_set(self) -> None:
        builder = _response_builder(_A_RESPONSE_TEMPLATE, FieldPath(_RECORDS_FIELD))
        response = builder.with_status_code(239).build()
        assert response.status_code == 239

    def test_given_pagination_with_strategy_when_build_then_apply_strategy(self) -> None:
        builder = _response_builder(
            {"has_more_pages": False} | _SOME_RECORDS,
            FieldPath(_RECORDS_FIELD),
            pagination_strategy=FieldUpdatePaginationStrategy(FieldPath("has_more_pages"), "yes more page")
        )

        response = builder.with_pagination().build()

        assert _body(response)["has_more_pages"] == "yes more page"

    def test_given_no_pagination_strategy_but_pagination_when_build_then_raise_error(self) -> None:
        builder = _response_builder(_A_RESPONSE_TEMPLATE, FieldPath(_RECORDS_FIELD))
        with pytest.raises(ValueError):
            builder.with_pagination()


class UtilMethodsTest(TestCase):
    def test_from_resource_file(self) -> None:
        template = find_template("test-resource", __file__)
        assert template == {"test-source template": "this is a template for test-resource"}

    def test_given_cwd_doesnt_have_unit_tests_as_parent_when_from_resource_file__then_raise_error(self) -> None:
        with pytest.raises(ValueError):
            find_template("test-resource", str(FilePath(__file__).parent.parent.parent.parent))

    def test_given_records_path_invalid_when_create_builders_from_resource_then_raise_exception(self) -> None:
        with pytest.raises(ValueError):
            create_record_builder(_A_RESPONSE_TEMPLATE, NestedPath(["invalid", "record", "path"]))



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import pkgutil
import sys
from dataclasses import InitVar, dataclass, field
from typing import Any, Mapping, Union

from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.schema.schema_loader import SchemaLoader
from airbyte_cdk.sources.declarative.types import Config
from airbyte_cdk.sources.utils.schema_helpers import ResourceSchemaLoader


def _default_file_path() -> str:
    # Schema files are always in "source_<connector_name>/schemas/<stream_name>.json
    # The connector's module name can be inferred by looking at the modules loaded and look for the one starting with source_
    source_modules = [
        k for k, v in sys.modules.items() if "source_" in k
    ]  # example: ['source_exchange_rates', 'source_exchange_rates.source']
    if source_modules:
        module = source_modules[0].split(".")[0]
        return f"./{module}/schemas/{{{{parameters['name']}}}}.json"

    # If we are not in a source_ module, the most likely scenario is we're processing a manifest from the connector builder
    # server which does not require a json schema to be defined.
    return "./{{parameters['name']}}.json"


@dataclass
class JsonFileSchemaLoader(ResourceSchemaLoader, SchemaLoader):
    """
    Loads the schema from a json file

    Attributes:
        file_path (Union[InterpolatedString, str]): The path to the json file describing the schema
        name (str): The stream's name
        config (Config): The user-provided configuration as specified by the source's spec
        parameters (Mapping[str, Any]): Additional arguments to pass to the string interpolation if needed
    """

    config: Config
    parameters: InitVar[Mapping[str, Any]]
    file_path: Union[InterpolatedString, str] = field(default=None)

    def __post_init__(self, parameters: Mapping[str, Any]):
        if not self.file_path:
            self.file_path = _default_file_path()
        self.file_path = InterpolatedString.create(self.file_path, parameters=parameters)

    def get_json_schema(self) -> Mapping[str, Any]:
        # todo: It is worth revisiting if we can replace file_path with just file_name if every schema is in the /schemas directory
        # this would require that we find a creative solution to store or retrieve source_name in here since the files are mounted there
        json_schema_path = self._get_json_filepath()
        resource, schema_path = self.extract_resource_and_schema_path(json_schema_path)
        raw_json_file = pkgutil.get_data(resource, schema_path)

        if not raw_json_file:
            raise IOError(f"Cannot find file {json_schema_path}")
        try:
            raw_schema = json.loads(raw_json_file)
        except ValueError as err:
            raise RuntimeError(f"Invalid JSON file format for file {json_schema_path}") from err
        self.package_name = resource
        return self._resolve_schema_references(raw_schema)

    def _get_json_filepath(self):
        return self.file_path.eval(self.config)

    @staticmethod
    def extract_resource_and_schema_path(json_schema_path: str) -> (str, str):
        """
        When the connector is running on a docker container, package_data is accessible from the resource (source_<name>), so we extract
        the resource from the first part of the schema path and the remaining path is used to find the schema file. This is a slight
        hack to identify the source name while we are in the airbyte_cdk module.
        :param json_schema_path: The path to the schema JSON file
        :return: Tuple of the resource name and the path to the schema file
        """
        split_path = json_schema_path.split("/")

        if split_path[0] == "" or split_path[0] == ".":
            split_path = split_path[1:]

        if len(split_path) == 0:
            return "", ""

        if len(split_path) == 1:
            return "", split_path[0]

        return split_path[0], "/".join(split_path[1:])



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import unittest
from queue import Queue
from typing import Callable, Iterable, List
from unittest.mock import Mock

import pytest
from airbyte_cdk.sources.streams.concurrent.partition_reader import PartitionReader
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_cdk.sources.streams.concurrent.partitions.types import PartitionCompleteSentinel, QueueItem

_RECORDS = [
    Record({"id": 1, "name": "Jack"}, "stream"),
    Record({"id": 2, "name": "John"}, "stream"),
]


class PartitionReaderTest(unittest.TestCase):
    def setUp(self) -> None:
        self._queue: Queue[QueueItem] = Queue()
        self._partition_reader = PartitionReader(self._queue)

    def test_given_no_records_when_process_partition_then_only_emit_sentinel(self):
        self._partition_reader.process_partition(self._a_partition([]))

        while queue_item := self._queue.get():
            if not isinstance(queue_item, PartitionCompleteSentinel):
                pytest.fail("Only one PartitionCompleteSentinel is expected")
            break

    def test_given_read_partition_successful_when_process_partition_then_queue_records_and_sentinel(self):
        self._partition_reader.process_partition(self._a_partition(_RECORDS))

        actual_records = []
        while queue_item := self._queue.get():
            if isinstance(queue_item, PartitionCompleteSentinel):
                break
            actual_records.append(queue_item)

        assert _RECORDS == actual_records

    def test_given_exception_when_process_partition_then_queue_records_and_raise_exception(self):
        partition = Mock()
        exception = ValueError()
        partition.read.side_effect = self._read_with_exception(_RECORDS, exception)

        self._partition_reader.process_partition(partition)

        for i in range(len(_RECORDS)):
            assert self._queue.get() == _RECORDS[i]
        assert self._queue.get() == exception

    def _a_partition(self, records: List[Record]) -> Partition:
        partition = Mock(spec=Partition)
        partition.read.return_value = iter(records)
        return partition

    @staticmethod
    def _read_with_exception(records: List[Record], exception: Exception) -> Callable[[], Iterable[Record]]:
        def mocked_function() -> Iterable[Record]:
            yield from records
            raise exception
        return mocked_function



from .abstract_concurrent_file_based_cursor import AbstractConcurrentFileBasedCursor
from .file_based_noop_cursor import FileBasedNoopCursor
from .file_based_concurrent_cursor import FileBasedConcurrentCursor

__all__ = ["AbstractConcurrentFileBasedCursor", "FileBasedConcurrentCursor", "FileBasedNoopCursor"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from copy import deepcopy
from typing import Any, Mapping, Optional, Type

from airbyte_cdk.sources.file_based.availability_strategy.abstract_file_based_availability_strategy import (
    AbstractFileBasedAvailabilityStrategy,
)
from airbyte_cdk.sources.file_based.discovery_policy import AbstractDiscoveryPolicy, DefaultDiscoveryPolicy
from airbyte_cdk.sources.file_based.file_based_source import default_parsers
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader
from airbyte_cdk.sources.file_based.file_types.file_type_parser import FileTypeParser
from airbyte_cdk.sources.file_based.schema_validation_policies import AbstractSchemaValidationPolicy
from airbyte_cdk.sources.file_based.stream.cursor import AbstractFileBasedCursor
from airbyte_cdk.sources.source import TState
from unit_tests.sources.file_based.in_memory_files_source import InMemoryFilesSource
from unit_tests.sources.file_based.scenarios.scenario_builder import SourceBuilder


class FileBasedSourceBuilder(SourceBuilder[InMemoryFilesSource]):
    def __init__(self) -> None:
        self._files: Mapping[str, Any] = {}
        self._file_type: Optional[str] = None
        self._availability_strategy: Optional[AbstractFileBasedAvailabilityStrategy] = None
        self._discovery_policy: AbstractDiscoveryPolicy = DefaultDiscoveryPolicy()
        self._validation_policies: Optional[Mapping[str, AbstractSchemaValidationPolicy]] = None
        self._parsers = default_parsers
        self._stream_reader: Optional[AbstractFileBasedStreamReader] = None
        self._file_write_options: Mapping[str, Any] = {}
        self._cursor_cls: Optional[Type[AbstractFileBasedCursor]] = None
        self._config: Optional[Mapping[str, Any]] = None
        self._state: Optional[TState] = None

    def build(self, configured_catalog: Optional[Mapping[str, Any]], config: Optional[Mapping[str, Any]], state: Optional[TState]) -> InMemoryFilesSource:
        if self._file_type is None:
            raise ValueError("file_type is not set")
        return InMemoryFilesSource(
            self._files,
            self._file_type,
            self._availability_strategy,
            self._discovery_policy,
            self._validation_policies,
            self._parsers,
            self._stream_reader,
            configured_catalog,
            config,
            state,
            self._file_write_options,
            self._cursor_cls,
        )

    def set_files(self, files: Mapping[str, Any]) -> "FileBasedSourceBuilder":
        self._files = files
        return self

    def set_file_type(self, file_type: str) -> "FileBasedSourceBuilder":
        self._file_type = file_type
        return self

    def set_parsers(self, parsers: Mapping[Type[Any], FileTypeParser]) -> "FileBasedSourceBuilder":
        self._parsers = parsers
        return self

    def set_availability_strategy(self, availability_strategy: AbstractFileBasedAvailabilityStrategy) -> "FileBasedSourceBuilder":
        self._availability_strategy = availability_strategy
        return self

    def set_discovery_policy(self, discovery_policy: AbstractDiscoveryPolicy) -> "FileBasedSourceBuilder":
        self._discovery_policy = discovery_policy
        return self

    def set_validation_policies(self, validation_policies: Mapping[str, AbstractSchemaValidationPolicy]) -> "FileBasedSourceBuilder":
        self._validation_policies = validation_policies
        return self

    def set_stream_reader(self, stream_reader: AbstractFileBasedStreamReader) -> "FileBasedSourceBuilder":
        self._stream_reader = stream_reader
        return self

    def set_cursor_cls(self, cursor_cls: AbstractFileBasedCursor) -> "FileBasedSourceBuilder":
        self._cursor_cls = cursor_cls
        return self

    def set_file_write_options(self, file_write_options: Mapping[str, Any]) -> "FileBasedSourceBuilder":
        self._file_write_options = file_write_options
        return self

    def copy(self) -> "FileBasedSourceBuilder":
        return deepcopy(self)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from unittest.mock import MagicMock, Mock, patch

import pytest
import requests
from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, Level, SyncMode, Type
from airbyte_cdk.sources.declarative.auth.declarative_authenticator import NoAuth
from airbyte_cdk.sources.declarative.incremental import Cursor, DatetimeBasedCursor
from airbyte_cdk.sources.declarative.partition_routers import SinglePartitionRouter
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_status import ResponseStatus
from airbyte_cdk.sources.declarative.requesters.request_option import RequestOptionType
from airbyte_cdk.sources.declarative.requesters.requester import HttpMethod
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever, SimpleRetrieverTestReadDecorator
from airbyte_cdk.sources.declarative.types import Record

A_SLICE_STATE = {"slice_state": "slice state value"}
A_STREAM_SLICE = {"stream slice": "slice value"}
A_STREAM_STATE = {"stream state": "state value"}

primary_key = "pk"
records = [{"id": 1}, {"id": 2}]
request_response_logs = [
    AirbyteLogMessage(level=Level.INFO, message="request:{}"),
    AirbyteLogMessage(level=Level.INFO, message="response{}"),
]
config = {}


@patch.object(SimpleRetriever, "_read_pages", return_value=iter([]))
def test_simple_retriever_full(mock_http_stream):
    requester = MagicMock()
    request_params = {"param": "value"}
    requester.get_request_params.return_value = request_params

    paginator = MagicMock()
    next_page_token = {"cursor": "cursor_value"}
    paginator.path.return_value = None
    paginator.next_page_token.return_value = next_page_token
    paginator.get_requesyyt_headers.return_value = {}

    record_selector = MagicMock()
    record_selector.select_records.return_value = records

    cursor = MagicMock(spec=Cursor)
    stream_slices = [{"date": "2022-01-01"}, {"date": "2022-01-02"}]
    cursor.stream_slices.return_value = stream_slices

    response = requests.Response()
    response.status_code = 200

    underlying_state = {"date": "2021-01-01"}
    cursor.get_stream_state.return_value = underlying_state

    requester.get_authenticator.return_value = NoAuth({})
    url_base = "https://airbyte.io"
    requester.get_url_base.return_value = url_base
    path = "/v1"
    requester.get_path.return_value = path
    http_method = HttpMethod.GET
    requester.get_method.return_value = http_method
    backoff_time = 60
    should_retry = ResponseStatus.retry(backoff_time)
    requester.interpret_response_status.return_value = should_retry
    request_body_json = {"body": "json"}
    requester.request_body_json.return_value = request_body_json

    request_body_data = {"body": "data"}
    requester.get_request_body_data.return_value = request_body_data
    request_body_json = {"body": "json"}
    requester.get_request_body_json.return_value = request_body_json
    request_kwargs = {"kwarg": "value"}
    requester.request_kwargs.return_value = request_kwargs

    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        paginator=paginator,
        record_selector=record_selector,
        stream_slicer=cursor,
        cursor=cursor,
        parameters={},
        config={},
    )

    assert retriever.primary_key == primary_key
    assert retriever.state == underlying_state
    assert retriever._next_page_token(response) == next_page_token
    assert retriever._request_params(None, None, None) == {}
    assert retriever.stream_slices() == stream_slices

    assert retriever._last_response is None
    assert retriever._records_from_last_response == []
    assert retriever._parse_response(response, stream_state={}, records_schema={}) == records
    assert retriever._last_response == response
    assert retriever._records_from_last_response == records

    [r for r in retriever.read_records(SyncMode.full_refresh)]
    paginator.reset.assert_called()


@patch.object(SimpleRetriever, "_read_pages", return_value=iter([*request_response_logs, *records]))
def test_simple_retriever_with_request_response_logs(mock_http_stream):
    requester = MagicMock()
    paginator = MagicMock()
    record_selector = MagicMock()
    stream_slicer = DatetimeBasedCursor(
        start_datetime="",
        end_datetime="",
        step="P1D",
        cursor_field="id",
        datetime_format="",
        cursor_granularity="P1D",
        config={},
        parameters={},
    )

    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        paginator=paginator,
        record_selector=record_selector,
        stream_slicer=stream_slicer,
        parameters={},
        config={},
    )

    actual_messages = [r for r in retriever.read_records(SyncMode.full_refresh)]
    paginator.reset.assert_called()

    assert isinstance(actual_messages[0], AirbyteLogMessage)
    assert isinstance(actual_messages[1], AirbyteLogMessage)
    assert actual_messages[2] == records[0]
    assert actual_messages[3] == records[1]


@patch.object(SimpleRetriever, "_read_pages", return_value=iter([]))
def test_simple_retriever_with_request_response_log_last_records(mock_http_stream):
    requester = MagicMock()
    paginator = MagicMock()
    record_selector = MagicMock()
    record_selector.select_records.return_value = request_response_logs
    response = requests.Response()
    response.status_code = 200
    stream_slicer = DatetimeBasedCursor(
        start_datetime="",
        end_datetime="",
        step="P1D",
        cursor_field="id",
        datetime_format="",
        cursor_granularity="P1D",
        config={},
        parameters={},
    )

    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        paginator=paginator,
        record_selector=record_selector,
        stream_slicer=stream_slicer,
        parameters={},
        config={},
    )

    assert retriever._last_response is None
    assert retriever._records_from_last_response == []
    assert retriever._parse_response(response, stream_state={}, records_schema={}) == request_response_logs
    assert retriever._last_response == response
    assert retriever._records_from_last_response == request_response_logs

    [r for r in retriever.read_records(SyncMode.full_refresh)]
    paginator.reset.assert_called()


@pytest.mark.parametrize(
    "test_name, paginator_mapping, stream_slicer_mapping, expected_mapping",
    [
        ("test_empty_headers", {}, {}, {}),
        ("test_header_from_pagination_and_slicer", {"offset": 1000}, {"key": "value"}, {"key": "value", "offset": 1000}),
        ("test_header_from_stream_slicer", {}, {"slice": "slice_value"}, {"slice": "slice_value"}),
        ("test_duplicate_header_slicer_paginator", {"k": "v"}, {"k": "slice_value"}, None),
    ],
)
def test_get_request_options_from_pagination(test_name, paginator_mapping, stream_slicer_mapping, expected_mapping):
    # This test does not test request headers because they must be strings
    paginator = MagicMock()
    paginator.get_request_params.return_value = paginator_mapping
    paginator.get_request_body_data.return_value = paginator_mapping
    paginator.get_request_body_json.return_value = paginator_mapping

    stream_slicer = MagicMock()
    stream_slicer.get_request_params.return_value = stream_slicer_mapping
    stream_slicer.get_request_body_data.return_value = stream_slicer_mapping
    stream_slicer.get_request_body_json.return_value = stream_slicer_mapping

    record_selector = MagicMock()
    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=MagicMock(),
        record_selector=record_selector,
        paginator=paginator,
        stream_slicer=stream_slicer,
        parameters={},
        config={},
    )

    request_option_type_to_method = {
        RequestOptionType.request_parameter: retriever._request_params,
        RequestOptionType.body_data: retriever._request_body_data,
        RequestOptionType.body_json: retriever._request_body_json,
    }

    for _, method in request_option_type_to_method.items():
        if expected_mapping is not None:
            actual_mapping = method(None, None, None)
            assert expected_mapping == actual_mapping
        else:
            try:
                method(None, None, None)
                assert False
            except ValueError:
                pass


@pytest.mark.parametrize(
    "test_name, paginator_mapping, expected_mapping",
    [
        ("test_only_base_headers", {}, {"key": "value"}),
        ("test_header_from_pagination", {"offset": 1000}, {"key": "value", "offset": "1000"}),
        ("test_duplicate_header", {"key": 1000}, None),
    ],
)
def test_get_request_headers(test_name, paginator_mapping, expected_mapping):
    # This test is separate from the other request options because request headers must be strings
    paginator = MagicMock()
    paginator.get_request_headers.return_value = paginator_mapping
    requester = MagicMock(use_cache=False)

    stream_slicer = MagicMock()
    stream_slicer.get_request_headers.return_value = {"key": "value"}

    record_selector = MagicMock()
    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        record_selector=record_selector,
        stream_slicer=stream_slicer,
        paginator=paginator,
        parameters={},
        config={},
    )

    request_option_type_to_method = {
        RequestOptionType.header: retriever._request_headers,
    }

    for _, method in request_option_type_to_method.items():
        if expected_mapping:
            actual_mapping = method(None, None, None)
            assert expected_mapping == actual_mapping
        else:
            try:
                method(None, None, None)
                assert False
            except ValueError:
                pass


@pytest.mark.parametrize(
    "test_name, slicer_body_data, paginator_body_data, expected_body_data",
    [
        ("test_only_slicer_mapping", {"key": "value"}, {}, {"key": "value"}),
        ("test_only_slicer_string", "key=value", {}, "key=value"),
        ("test_slicer_mapping_and_paginator_no_duplicate", {"key": "value"}, {"offset": 1000}, {"key": "value", "offset": 1000}),
        ("test_slicer_mapping_and_paginator_with_duplicate", {"key": "value"}, {"key": 1000}, None),
        ("test_slicer_string_and_paginator", "key=value", {"offset": 1000}, None),
    ],
)
def test_request_body_data(test_name, slicer_body_data, paginator_body_data, expected_body_data):
    paginator = MagicMock()
    paginator.get_request_body_data.return_value = paginator_body_data
    requester = MagicMock(use_cache=False)

    stream_slicer = MagicMock()
    stream_slicer.get_request_body_data.return_value = slicer_body_data

    record_selector = MagicMock()
    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        record_selector=record_selector,
        paginator=paginator,
        stream_slicer=stream_slicer,
        parameters={},
        config={},
    )

    if expected_body_data:
        actual_body_data = retriever._request_body_data(None, None, None)
        assert expected_body_data == actual_body_data
    else:
        try:
            retriever._request_body_data(None, None, None)
            assert False
        except ValueError:
            pass


@pytest.mark.parametrize(
    "test_name, requester_path, paginator_path, expected_path",
    [
        ("test_path_from_requester", "/v1/path", None, None),
        ("test_path_from_paginator", "/v1/path/", "/v2/paginator", "/v2/paginator"),
    ],
)
def test_path(test_name, requester_path, paginator_path, expected_path):
    paginator = MagicMock()
    paginator.path.return_value = paginator_path
    requester = MagicMock(use_cache=False)

    requester.get_path.return_value = requester_path

    record_selector = MagicMock()
    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        record_selector=record_selector,
        paginator=paginator,
        parameters={},
        config={},
    )

    actual_path = retriever._paginator_path()
    assert expected_path == actual_path


def test_limit_stream_slices():
    maximum_number_of_slices = 4
    stream_slicer = MagicMock()
    stream_slicer.stream_slices.return_value = _generate_slices(maximum_number_of_slices * 2)
    retriever = SimpleRetrieverTestReadDecorator(
        name="stream_name",
        primary_key=primary_key,
        requester=MagicMock(),
        paginator=MagicMock(),
        record_selector=MagicMock(),
        stream_slicer=stream_slicer,
        maximum_number_of_slices=maximum_number_of_slices,
        parameters={},
        config={},
    )

    truncated_slices = list(retriever.stream_slices())

    assert truncated_slices == _generate_slices(maximum_number_of_slices)


@pytest.mark.parametrize(
    "test_name, first_greater_than_second",
    [
        ("test_first_greater_than_second", True),
        ("test_second_greater_than_first", False),
    ],
)
def test_when_read_records_then_cursor_close_slice_with_greater_record(test_name, first_greater_than_second):
    first_record = Record({"first": 1}, {})
    second_record = Record({"second": 2}, {})
    records = [first_record, second_record]
    record_selector = MagicMock()
    record_selector.select_records.return_value = records
    cursor = MagicMock(spec=Cursor)
    cursor.is_greater_than_or_equal.return_value = first_greater_than_second
    paginator = MagicMock()
    paginator.get_request_headers.return_value = {}

    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=MagicMock(),
        paginator=paginator,
        record_selector=record_selector,
        stream_slicer=cursor,
        cursor=cursor,
        parameters={},
        config={},
    )
    stream_slice = {"repository": "airbyte"}

    def retriever_read_pages(_, __, ___):
        return retriever._parse_records(response=MagicMock(), stream_state={}, stream_slice=stream_slice, records_schema={})

    with patch.object(
        SimpleRetriever,
        "_read_pages",
        return_value=iter([first_record, second_record]),
        side_effect=retriever_read_pages,
    ):
        list(retriever.read_records(stream_slice=stream_slice, records_schema={}))
        cursor.close_slice.assert_called_once_with(stream_slice, first_record if first_greater_than_second else second_record)


def test_given_stream_data_is_not_record_when_read_records_then_update_slice_with_optional_record():
    stream_data = [AirbyteMessage(type=Type.LOG, log=AirbyteLogMessage(level=Level.INFO, message="a log message"))]
    record_selector = MagicMock()
    record_selector.select_records.return_value = []
    cursor = MagicMock(spec=Cursor)

    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=MagicMock(),
        paginator=Mock(),
        record_selector=record_selector,
        stream_slicer=cursor,
        cursor=cursor,
        parameters={},
        config={},
    )
    stream_slice = {"repository": "airbyte"}

    def retriever_read_pages(_, __, ___):
        return retriever._parse_records(response=MagicMock(), stream_state={}, stream_slice=stream_slice, records_schema={})

    with patch.object(
        SimpleRetriever,
        "_read_pages",
        return_value=iter(stream_data),
        side_effect=retriever_read_pages,
    ):
        list(retriever.read_records(stream_slice=stream_slice, records_schema={}))
        cursor.close_slice.assert_called_once_with(stream_slice, None)


def _generate_slices(number_of_slices):
    return [{"date": f"2022-01-0{day + 1}"} for day in range(number_of_slices)]


@patch.object(SimpleRetriever, "_read_pages", return_value=iter([]))
def test_given_state_selector_when_read_records_use_stream_state(http_stream_read_pages, mocker):
    requester = MagicMock()
    paginator = MagicMock()
    record_selector = MagicMock()
    cursor = MagicMock(spec=Cursor)
    cursor.select_state = MagicMock(return_value=A_SLICE_STATE)
    cursor.get_stream_state = MagicMock(return_value=A_STREAM_STATE)

    retriever = SimpleRetriever(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        paginator=paginator,
        record_selector=record_selector,
        stream_slicer=cursor,
        cursor=cursor,
        parameters={},
        config={},
    )

    list(retriever.read_records(stream_slice=A_STREAM_SLICE, records_schema={}))

    http_stream_read_pages.assert_called_once_with(mocker.ANY, A_STREAM_STATE, A_STREAM_SLICE)


def test_emit_log_request_response_messages(mocker):
    record_selector = MagicMock()
    record_selector.select_records.return_value = records

    request = requests.PreparedRequest()
    request.headers = {"header": "value"}
    request.url = "http://byrde.enterprises.com/casinos"

    response = requests.Response()
    response.request = request
    response.status_code = 200

    format_http_message_mock = mocker.patch("airbyte_cdk.sources.declarative.retrievers.simple_retriever.format_http_message")
    requester = MagicMock()
    retriever = SimpleRetrieverTestReadDecorator(
        name="stream_name",
        primary_key=primary_key,
        requester=requester,
        paginator=MagicMock(),
        record_selector=record_selector,
        stream_slicer=SinglePartitionRouter(parameters={}),
        parameters={},
        config={},
    )

    retriever._fetch_next_page(stream_state={}, stream_slice={})

    assert requester.send_request.call_args_list[0][1]["log_formatter"] is not None
    assert requester.send_request.call_args_list[0][1]["log_formatter"](response) == format_http_message_mock.return_value



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, Mapping, Optional

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.streams.core import Stream, StreamData


def get_first_stream_slice(stream) -> Optional[Mapping[str, Any]]:
    """
    Gets the first stream_slice from a given stream's stream_slices.
    :param stream: stream
    :raises StopIteration: if there is no first slice to return (the stream_slices generator is empty)
    :return: first stream slice from 'stream_slices' generator (`None` is a valid stream slice)
    """
    # We wrap the return output of stream_slices() because some implementations return types that are iterable,
    # but not iterators such as lists or tuples
    slices = iter(
        stream.stream_slices(
            cursor_field=stream.cursor_field,
            sync_mode=SyncMode.full_refresh,
        )
    )
    return next(slices)


def get_first_record_for_slice(stream: Stream, stream_slice: Optional[Mapping[str, Any]]) -> StreamData:
    """
    Gets the first record for a stream_slice of a stream.
    :param stream: stream
    :param stream_slice: stream_slice
    :raises StopIteration: if there is no first record to return (the read_records generator is empty)
    :return: StreamData containing the first record in the slice
    """
    # We wrap the return output of read_records() because some implementations return types that are iterable,
    # but not iterators such as lists or tuples
    records_for_slice = iter(stream.read_records(sync_mode=SyncMode.full_refresh, stream_slice=stream_slice))
    return next(records_for_slice)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from typing import Any

from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream


class PartitionGenerationCompletedSentinel:
    """
    A sentinel object indicating all partitions for a stream were produced.
    Includes a pointer to the stream that was processed.
    """

    def __init__(self, stream: AbstractStream):
        """
        :param stream: The stream that was processed
        """
        self.stream = stream

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PartitionGenerationCompletedSentinel):
            return self.stream == other.stream
        return False



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import unittest
from unittest.mock import Mock

from airbyte_cdk.models import AirbyteStream, SyncMode
from airbyte_cdk.sources.streams.concurrent.availability_strategy import STREAM_AVAILABLE
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor, NoopCursor
from airbyte_cdk.sources.streams.concurrent.default_stream import DefaultStream


class ThreadBasedConcurrentStreamTest(unittest.TestCase):
    def setUp(self):
        self._partition_generator = Mock()
        self._name = "name"
        self._json_schema = {}
        self._availability_strategy = Mock()
        self._primary_key = []
        self._cursor_field = None
        self._logger = Mock()
        self._cursor = Mock(spec=Cursor)
        self._stream = DefaultStream(
            self._partition_generator,
            self._name,
            self._json_schema,
            self._availability_strategy,
            self._primary_key,
            self._cursor_field,
            self._logger,
            NoopCursor(),
        )

    def test_get_json_schema(self):
        json_schema = self._stream.get_json_schema()
        assert json_schema == self._json_schema

    def test_check_availability(self):
        self._availability_strategy.check_availability.return_value = STREAM_AVAILABLE
        availability = self._stream.check_availability()
        assert availability == STREAM_AVAILABLE
        self._availability_strategy.check_availability.assert_called_once_with(self._logger)

    def test_check_for_error_raises_an_exception_if_any_of_the_futures_are_not_done(self):
        futures = [Mock() for _ in range(3)]
        for f in futures:
            f.exception.return_value = None
        futures[0].done.return_value = False

        with self.assertRaises(Exception):
            self._stream._check_for_errors(futures)

    def test_check_for_error_raises_an_exception_if_any_of_the_futures_raised_an_exception(self):
        futures = [Mock() for _ in range(3)]
        for f in futures:
            f.exception.return_value = None
        futures[0].exception.return_value = Exception("error")

        with self.assertRaises(Exception):
            self._stream._check_for_errors(futures)

    def test_as_airbyte_stream(self):
        expected_airbyte_stream = AirbyteStream(
            name=self._name,
            json_schema=self._json_schema,
            supported_sync_modes=[SyncMode.full_refresh],
            source_defined_cursor=None,
            default_cursor_field=None,
            source_defined_primary_key=None,
            namespace=None,
        )
        actual_airbyte_stream = self._stream.as_airbyte_stream()

        assert expected_airbyte_stream == actual_airbyte_stream

    def test_as_airbyte_stream_with_primary_key(self):
        json_schema = {
            "type": "object",
            "properties": {
                "id_a": {"type": ["null", "string"]},
                "id_b": {"type": ["null", "string"]},
            },
        }
        stream = DefaultStream(
            self._partition_generator,
            self._name,
            json_schema,
            self._availability_strategy,
            ["id"],
            self._cursor_field,
            self._logger,
            NoopCursor(),
        )

        expected_airbyte_stream = AirbyteStream(
            name=self._name,
            json_schema=json_schema,
            supported_sync_modes=[SyncMode.full_refresh],
            source_defined_cursor=None,
            default_cursor_field=None,
            source_defined_primary_key=[["id"]],
            namespace=None,
        )

        airbyte_stream = stream.as_airbyte_stream()
        assert expected_airbyte_stream == airbyte_stream

    def test_as_airbyte_stream_with_composite_primary_key(self):
        json_schema = {
            "type": "object",
            "properties": {
                "id_a": {"type": ["null", "string"]},
                "id_b": {"type": ["null", "string"]},
            },
        }
        stream = DefaultStream(
            self._partition_generator,
            self._name,
            json_schema,
            self._availability_strategy,
            ["id_a", "id_b"],
            self._cursor_field,
            self._logger,
            NoopCursor(),
        )

        expected_airbyte_stream = AirbyteStream(
            name=self._name,
            json_schema=json_schema,
            supported_sync_modes=[SyncMode.full_refresh],
            source_defined_cursor=None,
            default_cursor_field=None,
            source_defined_primary_key=[["id_a", "id_b"]],
            namespace=None,
        )

        airbyte_stream = stream.as_airbyte_stream()
        assert expected_airbyte_stream == airbyte_stream

    def test_as_airbyte_stream_with_a_cursor(self):
        json_schema = {
            "type": "object",
            "properties": {
                "id": {"type": ["null", "string"]},
                "date": {"type": ["null", "string"]},
            },
        }
        stream = DefaultStream(
            self._partition_generator,
            self._name,
            json_schema,
            self._availability_strategy,
            self._primary_key,
            "date",
            self._logger,
            NoopCursor(),
        )

        expected_airbyte_stream = AirbyteStream(
            name=self._name,
            json_schema=json_schema,
            supported_sync_modes=[SyncMode.full_refresh, SyncMode.incremental],
            source_defined_cursor=True,
            default_cursor_field=["date"],
            source_defined_primary_key=None,
            namespace=None,
        )

        airbyte_stream = stream.as_airbyte_stream()
        assert expected_airbyte_stream == airbyte_stream

    def test_as_airbyte_stream_with_namespace(self):
        stream = DefaultStream(
            self._partition_generator,
            self._name,
            self._json_schema,
            self._availability_strategy,
            self._primary_key,
            self._cursor_field,
            self._logger,
            NoopCursor(),
            namespace="test",
        )
        expected_airbyte_stream = AirbyteStream(
            name=self._name,
            json_schema=self._json_schema,
            supported_sync_modes=[SyncMode.full_refresh],
            source_defined_cursor=None,
            default_cursor_field=None,
            source_defined_primary_key=None,
            namespace="test",
        )
        actual_airbyte_stream = stream.as_airbyte_stream()

        assert expected_airbyte_stream == actual_airbyte_stream



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from unittest import TestCase

import pytest
import requests
from airbyte_cdk.test.mock_http import HttpMocker, HttpRequest, HttpResponse

# Ensure that the scheme is HTTP as requests only partially supports other schemes
# see https://github.com/psf/requests/blob/0b4d494192de489701d3a2e32acef8fb5d3f042e/src/requests/models.py#L424-L429
_A_URL = "http://test.com/"
_ANOTHER_URL = "http://another-test.com/"
_A_RESPONSE_BODY = "a body"
_ANOTHER_RESPONSE_BODY = "another body"
_A_RESPONSE = HttpResponse("any response")
_SOME_QUERY_PARAMS = {"q1": "query value"}
_SOME_HEADERS = {"h1": "header value"}
_OTHER_HEADERS = {"h2": "another header value"}
_SOME_REQUEST_BODY_MAPPING = {"first_field": "first_value", "second_field": 2}
_SOME_REQUEST_BODY_STR = "some_request_body"


class HttpMockerTest(TestCase):
    @HttpMocker()
    def test_given_get_request_match_when_decorate_then_return_response(self, http_mocker):
        http_mocker.get(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS),
            HttpResponse(_A_RESPONSE_BODY, 474, _OTHER_HEADERS),
        )

        response = requests.get(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS)

        assert response.text == _A_RESPONSE_BODY
        assert response.status_code == 474
        assert response.headers == _OTHER_HEADERS

    @HttpMocker()
    def test_given_loose_headers_matching_when_decorate_then_match(self, http_mocker):
        http_mocker.get(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS),
            HttpResponse(_A_RESPONSE_BODY, 474),
        )

        requests.get(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS | {"more strict query param key": "any value"})

    @HttpMocker()
    def test_given_post_request_match_when_decorate_then_return_response(self, http_mocker):
        http_mocker.post(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS, _SOME_REQUEST_BODY_STR),
            HttpResponse(_A_RESPONSE_BODY, 474),
        )

        response = requests.post(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS, data=_SOME_REQUEST_BODY_STR)

        assert response.text == _A_RESPONSE_BODY
        assert response.status_code == 474

    @HttpMocker()
    def test_given_multiple_responses_when_decorate_get_request_then_return_response(self, http_mocker):
        http_mocker.get(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS),
            [HttpResponse(_A_RESPONSE_BODY, 1), HttpResponse(_ANOTHER_RESPONSE_BODY, 2)],
        )

        first_response = requests.get(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS)
        second_response = requests.get(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS)

        assert first_response.text == _A_RESPONSE_BODY
        assert first_response.status_code == 1
        assert second_response.text == _ANOTHER_RESPONSE_BODY
        assert second_response.status_code == 2

    @HttpMocker()
    def test_given_multiple_responses_when_decorate_post_request_then_return_response(self, http_mocker):
        http_mocker.post(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS, _SOME_REQUEST_BODY_STR),
            [HttpResponse(_A_RESPONSE_BODY, 1), HttpResponse(_ANOTHER_RESPONSE_BODY, 2)],
        )

        first_response = requests.post(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS, data=_SOME_REQUEST_BODY_STR)
        second_response = requests.post(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS, data=_SOME_REQUEST_BODY_STR)

        assert first_response.text == _A_RESPONSE_BODY
        assert first_response.status_code == 1
        assert second_response.text == _ANOTHER_RESPONSE_BODY
        assert second_response.status_code == 2

    @HttpMocker()
    def test_given_more_requests_than_responses_when_decorate_then_raise_error(self, http_mocker):
        http_mocker.get(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS),
            [HttpResponse(_A_RESPONSE_BODY, 1), HttpResponse(_ANOTHER_RESPONSE_BODY, 2)],
        )

        last_response = [requests.get(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS) for _ in range(10)][-1]

        assert last_response.text == _ANOTHER_RESPONSE_BODY
        assert last_response.status_code == 2

    @HttpMocker()
    def test_given_all_requests_match_when_decorate_then_do_not_raise(self, http_mocker):
        http_mocker.get(
            HttpRequest(_A_URL, _SOME_QUERY_PARAMS, _SOME_HEADERS),
            _A_RESPONSE,
        )
        requests.get(_A_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS)

    def test_given_missing_requests_when_decorate_then_raise(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            http_mocker.get(
                HttpRequest(_A_URL),
                _A_RESPONSE,
            )

        with pytest.raises(ValueError) as exc_info:
            decorated_function()
        assert "Invalid number of matches" in str(exc_info.value)

    def test_given_assertion_error_when_decorate_then_raise_assertion_error(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            http_mocker.get(
                HttpRequest(_A_URL),
                _A_RESPONSE,
            )
            requests.get(_A_URL)
            assert False

        with pytest.raises(AssertionError):
            decorated_function()

    def test_given_assertion_error_but_missing_request_when_decorate_then_raise_missing_http_request(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            http_mocker.get(
                HttpRequest(_A_URL),
                _A_RESPONSE,
            )
            assert False

        with pytest.raises(ValueError) as exc_info:
            decorated_function()
        assert "Invalid number of matches" in str(exc_info.value)

    def test_given_request_does_not_match_when_decorate_then_raise(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            http_mocker.get(
                HttpRequest(_A_URL),
                _A_RESPONSE,
            )
            requests.get(_ANOTHER_URL, params=_SOME_QUERY_PARAMS, headers=_SOME_HEADERS)

        with pytest.raises(ValueError) as exc_info:
            decorated_function()
        assert "No matcher matches" in str(exc_info.value)

    def test_given_request_matches_multiple_matchers_when_decorate_then_match_first_one(self):
        less_granular_headers = {"less_granular": "1"}
        more_granular_headers = {"more_granular": "2"} | less_granular_headers

        @HttpMocker()
        def decorated_function(http_mocker):
            http_mocker.get(
                HttpRequest(_A_URL, headers=more_granular_headers),
                _A_RESPONSE,
            )
            http_mocker.get(
                HttpRequest(_A_URL, headers=less_granular_headers),
                _A_RESPONSE,
            )
            requests.get(_A_URL, headers=more_granular_headers)

        with pytest.raises(ValueError) as exc_info:
            decorated_function()
        assert "more_granular" in str(exc_info.value)  # the matcher corresponding to the first `http_mocker.get` is not matched

    def test_given_exact_number_of_call_provided_when_assert_number_of_calls_then_do_not_raise(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            request = HttpRequest(_A_URL)
            http_mocker.get(request, _A_RESPONSE)

            requests.get(_A_URL)
            requests.get(_A_URL)

            http_mocker.assert_number_of_calls(request, 2)

        decorated_function()
        # then do not raise

    def test_given_invalid_number_of_call_provided_when_assert_number_of_calls_then_raise(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            request = HttpRequest(_A_URL)
            http_mocker.get(request, _A_RESPONSE)

            requests.get(_A_URL)
            requests.get(_A_URL)

            http_mocker.assert_number_of_calls(request, 1)

        with pytest.raises(AssertionError):
            decorated_function()

    def test_given_unknown_request_when_assert_number_of_calls_then_raise(self):
        @HttpMocker()
        def decorated_function(http_mocker):
            http_mocker.get(HttpRequest(_A_URL), _A_RESPONSE)
            http_mocker.assert_number_of_calls(HttpRequest(_ANOTHER_URL), 1)

        with pytest.raises(ValueError):
            decorated_function()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from dataclasses import dataclass
from typing import Union

import requests
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_status import ResponseStatus


@dataclass
class ErrorHandler:
    """
    Defines whether a request was successful and how to handle a failure.
    """

    @property
    @abstractmethod
    def max_retries(self) -> Union[int, None]:
        """
        Specifies maximum amount of retries for backoff policy. Return None for no limit.
        """
        pass

    @property
    @abstractmethod
    def max_time(self) -> Union[int, None]:
        """
        Specifies maximum total waiting time (in seconds) for backoff policy. Return None for no limit.
        """
        pass

    @abstractmethod
    def interpret_response(self, response: requests.Response) -> ResponseStatus:
        """
        Evaluate response status describing whether a failing request should be retried or ignored.

        :param response: response to evaluate
        :return: response status
        """
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass
from typing import Any, Dict, List, Optional


@dataclass
class HttpResponse:
    status: int
    body: Optional[str] = None
    headers: Optional[Dict[str, Any]] = None


@dataclass
class HttpRequest:
    url: str
    parameters: Optional[Dict[str, Any]]
    headers: Optional[Dict[str, Any]]
    http_method: str
    body: Optional[str] = None


@dataclass
class StreamReadPages:
    records: List[object]
    request: Optional[HttpRequest] = None
    response: Optional[HttpResponse] = None


@dataclass
class StreamReadSlices:
    pages: List[StreamReadPages]
    slice_descriptor: Optional[Dict[str, Any]]
    state: Optional[Dict[str, Any]] = None


@dataclass
class LogMessage:
    message: str
    level: str


@dataclass
class AuxiliaryRequest:
    title: str
    description: str
    request: HttpRequest
    response: HttpResponse


@dataclass
class StreamRead(object):
    logs: List[LogMessage]
    slices: List[StreamReadSlices]
    test_read_limit_reached: bool
    auxiliary_requests: List[AuxiliaryRequest]
    inferred_schema: Optional[Dict[str, Any]]
    inferred_datetime_formats: Optional[Dict[str, str]]
    latest_config_update: Optional[Dict[str, Any]]


@dataclass
class StreamReadRequestBody:
    manifest: Dict[str, Any]
    stream: str
    config: Dict[str, Any]
    state: Optional[Dict[str, Any]]
    record_limit: Optional[int]



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.


from datetime import datetime
from typing import Any, Dict, List, MutableMapping, Optional, Tuple
from unittest.mock import MagicMock

import pytest
from airbyte_cdk.models import AirbyteStateMessage, SyncMode
from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.stream.concurrent.adapters import FileBasedStreamPartition
from airbyte_cdk.sources.file_based.stream.concurrent.cursor import FileBasedConcurrentCursor
from airbyte_cdk.sources.streams.concurrent.cursor import CursorField
from freezegun import freeze_time

DATE_TIME_FORMAT = FileBasedConcurrentCursor.DATE_TIME_FORMAT
MOCK_DAYS_TO_SYNC_IF_HISTORY_IS_FULL = 3


def _make_cursor(input_state: Optional[MutableMapping[str, Any]]) -> FileBasedConcurrentCursor:
    stream = MagicMock()
    stream.name = "test"
    stream.namespace = None
    stream_config = MagicMock()
    stream_config.days_to_sync_if_history_is_full = MOCK_DAYS_TO_SYNC_IF_HISTORY_IS_FULL
    cursor = FileBasedConcurrentCursor(
        stream_config,
        stream.name,
        None,
        input_state,
        MagicMock(),
        ConnectorStateManager(
            stream_instance_map={stream.name: stream},
            state=[AirbyteStateMessage.parse_obj(input_state)] if input_state is not None else None,
        ),
        CursorField(FileBasedConcurrentCursor.CURSOR_FIELD),
    )
    return cursor


@pytest.mark.parametrize(
    "input_state, expected_cursor_value",
    [
        pytest.param({}, (datetime.min, ""), id="no-state-gives-min-cursor"),
        pytest.param({"history": {}}, (datetime.min, ""), id="missing-cursor-field-gives-min-cursor"),
        pytest.param(
            {
                "history": {"a.csv": "2021-01-01T00:00:00.000000Z"},
                "_ab_source_file_last_modified": "2021-01-01T00:00:00.000000Z_a.csv"
            },
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            id="cursor-value-matches-earliest-file",
        ),
        pytest.param(
            {
                "history": {"a.csv": "2021-01-01T00:00:00.000000Z"},
                "_ab_source_file_last_modified": "2020-01-01T00:00:00.000000Z_a.csv"
            },
            (datetime.strptime("2020-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            id="cursor-value-is-earlier",
        ),
        pytest.param(
            {
                "history": {"a.csv": "2022-01-01T00:00:00.000000Z"},
                "_ab_source_file_last_modified": "2021-01-01T00:00:00.000000Z_a.csv"
            },
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            id="cursor-value-is-later",
        ),
        pytest.param(
            {
                "history": {
                    "a.csv": "2021-01-01T00:00:00.000000Z",
                    "b.csv": "2021-01-02T00:00:00.000000Z",
                    "c.csv": "2021-01-03T00:00:00.000000Z"
                },
                "_ab_source_file_last_modified": "2021-01-04T00:00:00.000000Z_d.csv"
            },
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            id="cursor-not-earliest",
        ),
        pytest.param(
            {
                "history": {"b.csv": "2020-12-31T00:00:00.000000Z"},
                "_ab_source_file_last_modified": "2021-01-01T00:00:00.000000Z_a.csv"
            },
            (datetime.strptime("2020-12-31T00:00:00.000000Z", DATE_TIME_FORMAT), "b.csv"),
            id="state-with-cursor-and-earlier-history"
        ),
        pytest.param(
            {
                "history": {"b.csv": "2021-01-02T00:00:00.000000Z"},
                "_ab_source_file_last_modified": "2021-01-01T00:00:00.000000Z_a.csv"
            },
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            id="state-with-cursor-and-later-history"
        ),
    ]
)
def test_compute_prev_sync_cursor(input_state: MutableMapping[str, Any], expected_cursor_value: Tuple[datetime, str]):
    cursor = _make_cursor(input_state)
    assert cursor._compute_prev_sync_cursor(input_state) == expected_cursor_value


@pytest.mark.parametrize(
    "initial_state, pending_files, file_to_add, expected_history, expected_pending_files, expected_cursor_value",
    [
        pytest.param(
            {"history": {}},
            [("newfile.csv", "2021-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [],
            "2021-01-05T00:00:00.000000Z_newfile.csv",
            id="add-to-empty-history-single-pending-file",
        ),
        pytest.param(
            {"history": {}},
            [("newfile.csv", "2021-01-05T00:00:00.000000Z"), ("pending.csv", "2020-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [("pending.csv", "2020-01-05T00:00:00.000000Z")],
            "2020-01-05T00:00:00.000000Z_pending.csv",
            id="add-to-empty-history-pending-file-is-older",
        ),
        pytest.param(
            {"history": {}},
            [("newfile.csv", "2021-01-05T00:00:00.000000Z"), ("pending.csv", "2022-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [("pending.csv", "2022-01-05T00:00:00.000000Z")],
            "2022-01-05T00:00:00.000000Z_pending.csv",
            id="add-to-empty-history-pending-file-is-newer",
        ),
        pytest.param(
            {"history": {"existing.csv": "2021-01-04T00:00:00.000000Z"}},
            [("newfile.csv", "2021-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"existing.csv": "2021-01-04T00:00:00.000000Z", "newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [],
            "2021-01-05T00:00:00.000000Z_newfile.csv",
            id="add-to-nonempty-history-single-pending-file",
        ),
        pytest.param(
            {"history": {"existing.csv": "2021-01-04T00:00:00.000000Z"}},
            [("newfile.csv", "2021-01-05T00:00:00.000000Z"), ("pending.csv", "2020-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"existing.csv": "2021-01-04T00:00:00.000000Z", "newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [("pending.csv", "2020-01-05T00:00:00.000000Z")],
            "2020-01-05T00:00:00.000000Z_pending.csv",
            id="add-to-nonempty-history-pending-file-is-older",
        ),
        pytest.param(
            {"history": {"existing.csv": "2021-01-04T00:00:00.000000Z"}},
            [("newfile.csv", "2021-01-05T00:00:00.000000Z"), ("pending.csv", "2022-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"existing.csv": "2021-01-04T00:00:00.000000Z", "newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [("pending.csv", "2022-01-05T00:00:00.000000Z")],
            "2022-01-05T00:00:00.000000Z_pending.csv",
            id="add-to-nonempty-history-pending-file-is-newer",
        ),
    ]
)
def test_add_file(
    initial_state: MutableMapping[str, Any],
    pending_files: List[Tuple[str, str]],
    file_to_add: Tuple[str, str],
    expected_history: Dict[str, Any],
    expected_pending_files: List[Tuple[str, str]],
    expected_cursor_value: str,
):
    cursor = _make_cursor(initial_state)
    mock_message_repository = MagicMock()
    cursor._message_repository = mock_message_repository
    stream = MagicMock()

    cursor.set_pending_partitions([
        FileBasedStreamPartition(
            stream,
            {"files": [RemoteFile(uri=uri, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT))]},
            mock_message_repository,
            SyncMode.full_refresh,
            FileBasedConcurrentCursor.CURSOR_FIELD,
            initial_state,
            cursor
        ) for uri, timestamp in pending_files
    ])

    uri, timestamp = file_to_add
    cursor.add_file(RemoteFile(uri=uri, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT)))
    assert cursor._file_to_datetime_history == expected_history
    assert cursor._pending_files == {uri: RemoteFile(uri=uri, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT)) for uri, timestamp in expected_pending_files}
    assert mock_message_repository.emit_message.call_args_list[0].args[0].state.data["test"]["_ab_source_file_last_modified"] == expected_cursor_value


@pytest.mark.parametrize(
    "initial_state, pending_files, file_to_add, expected_history, expected_pending_files, expected_cursor_value",
    [
        pytest.param(
            {"history": {}},
            [],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [],
            "2021-01-05T00:00:00.000000Z_newfile.csv",
            id="add-to-empty-history-no-pending-files",
        ),
        pytest.param(
            {"history": {}},
            [("pending.csv", "2021-01-05T00:00:00.000000Z")],
            ("newfile.csv", "2021-01-05T00:00:00.000000Z"),
            {"newfile.csv": "2021-01-05T00:00:00.000000Z"},
            [("pending.csv", "2021-01-05T00:00:00.000000Z")],
            "2021-01-05T00:00:00.000000Z_pending.csv",
            id="add-to-empty-history-file-not-in-pending-files",
        ),
    ]
)
def test_add_file_invalid(
    initial_state: MutableMapping[str, Any],
    pending_files: List[Tuple[str, str]],
    file_to_add: Tuple[str, str],
    expected_history: Dict[str, Any],
    expected_pending_files: List[Tuple[str, str]],
    expected_cursor_value: str,
):
    cursor = _make_cursor(initial_state)
    cursor._pending_files = {uri: RemoteFile(uri=uri, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT)) for uri, timestamp in pending_files}
    mock_message_repository = MagicMock()
    cursor._message_repository = mock_message_repository

    uri, timestamp = file_to_add
    cursor.add_file(RemoteFile(uri=uri, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT)))
    assert cursor._file_to_datetime_history == expected_history
    assert cursor._pending_files == {uri: RemoteFile(uri=uri, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT)) for uri, timestamp in expected_pending_files}
    assert mock_message_repository.emit_message.call_args_list[0].args[0].log.level.value == "WARN"
    assert mock_message_repository.emit_message.call_args_list[1].args[0].state.data["test"]["_ab_source_file_last_modified"] == expected_cursor_value


@pytest.mark.parametrize(
    "input_state, pending_files, expected_cursor_value",
    [
        pytest.param({}, [], f"{datetime.min.strftime('%Y-%m-%dT%H:%M:%S.%fZ')}_", id="no-state-no-pending"),
        pytest.param(
            {"history": {"a.csv": "2021-01-01T00:00:00.000000Z"}},
            [],
            "2021-01-01T00:00:00.000000Z_a.csv",
            id="no-pending-with-history"
        ),
        pytest.param(
            {"history": {}},
            [("b.csv", "2021-01-02T00:00:00.000000Z")],
            "2021-01-02T00:00:00.000000Z_b.csv",
            id="pending-no-history"
        ),
        pytest.param(
            {"history": {"a.csv": "2022-01-01T00:00:00.000000Z"}},
            [("b.csv", "2021-01-02T00:00:00.000000Z")],
            "2021-01-01T00:00:00.000000Z_a.csv",
            id="with-pending-before-history"
        ),
        pytest.param(
            {"history": {"a.csv": "2021-01-01T00:00:00.000000Z"}},
            [("b.csv", "2022-01-02T00:00:00.000000Z")],
            "2022-01-01T00:00:00.000000Z_a.csv",
            id="with-pending-after-history"
        ),
    ]
)
def test_get_new_cursor_value(input_state: MutableMapping[str, Any], pending_files: List[Tuple[str, str]], expected_cursor_value: str):
    cursor = _make_cursor(input_state)
    pending_partitions = []
    for url, timestamp in pending_files:
        partition = MagicMock()
        partition.to_slice = lambda *args, **kwargs: {"files": [RemoteFile(uri=url, last_modified=datetime.strptime(timestamp, DATE_TIME_FORMAT))]}
        pending_partitions.append(partition)

    cursor.set_pending_partitions(pending_partitions)


@pytest.mark.parametrize(
    "all_files, history, is_history_full, prev_cursor_value, expected_files_to_sync",
    [
        pytest.param(
            [RemoteFile(uri="new.csv", last_modified=datetime.strptime("2021-01-03T00:00:00.000000Z", "%Y-%m-%dT%H:%M:%S.%fZ"))],
            {},
            False,
            (datetime.min, ""),
            ["new.csv"],
            id="empty-history-one-new-file"
        ),
        pytest.param(
            [RemoteFile(uri="a.csv", last_modified=datetime.strptime("2021-01-02T00:00:00.000000Z", "%Y-%m-%dT%H:%M:%S.%fZ"))],
            {"a.csv": "2021-01-01T00:00:00.000000Z"},
            False,
            (datetime.min, ""),
            ["a.csv"],
            id="non-empty-history-file-in-history-modified"
        ),
        pytest.param(
            [RemoteFile(uri="a.csv", last_modified=datetime.strptime("2021-01-01T00:00:00.000000Z", "%Y-%m-%dT%H:%M:%S.%fZ"))],
            {"a.csv": "2021-01-01T00:00:00.000000Z"},
            False,
            (datetime.min, ""),
            [],
            id="non-empty-history-file-in-history-not-modified"
        ),
    ]
)
def test_get_files_to_sync(all_files, history, is_history_full, prev_cursor_value, expected_files_to_sync):
    cursor = _make_cursor({})
    cursor._file_to_datetime_history = history
    cursor._prev_cursor_value = prev_cursor_value
    cursor._is_history_full = MagicMock(return_value=is_history_full)
    files_to_sync = list(cursor.get_files_to_sync(all_files, MagicMock()))
    assert [f.uri for f in files_to_sync] == expected_files_to_sync


@freeze_time("2023-06-16T00:00:00Z")
@pytest.mark.parametrize(
    "file_to_check, history, is_history_full, prev_cursor_value, sync_start, expected_should_sync",
    [
        pytest.param(
            RemoteFile(uri="new.csv", last_modified=datetime.strptime("2021-01-03T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            False,
            (datetime.min, ""),
            datetime.min,
            True,
            id="file-not-in-history-not-full-old-cursor"
        ),
        pytest.param(
            RemoteFile(uri="new.csv", last_modified=datetime.strptime("2021-01-03T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            False,
            (datetime.strptime("2024-01-02T00:00:00.000000Z", DATE_TIME_FORMAT), ""),
            datetime.min,
            True,
            id="file-not-in-history-not-full-new-cursor"
        ),
        pytest.param(
            RemoteFile(uri="a.csv", last_modified=datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {"a.csv": "2021-01-01T00:00:00.000000Z"},
            False,
            (datetime.min, ""),
            datetime.min,
            False,
            id="file-in-history-not-modified"
        ),
        pytest.param(
            RemoteFile(uri="a.csv", last_modified=datetime.strptime("2020-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {"a.csv": "2021-01-01T00:00:00.000000Z"},
            False,
            (datetime.min, ""),
            datetime.min,
            False,
            id="file-in-history-modified-before"
        ),
        pytest.param(
            RemoteFile(uri="a.csv", last_modified=datetime.strptime("2022-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {"a.csv": "2021-01-01T00:00:00.000000Z"},
            False,
            (datetime.min, ""),
            datetime.min,
            True,
            id="file-in-history-modified-after"
        ),
        pytest.param(
            RemoteFile(uri="new.csv", last_modified=datetime.strptime("2022-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            True,
            (datetime.strptime("2021-01-02T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            datetime.min,
            True,
            id="history-full-file-modified-after-cursor"
        ),
        pytest.param(
            RemoteFile(uri="new1.csv", last_modified=datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            True,
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "new0.csv"),
            datetime.min,
            True,
            id="history-full-modified-eq-cursor-uri-gt"
        ),
        pytest.param(
            RemoteFile(uri="new0.csv", last_modified=datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            True,
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "new1.csv"),
            datetime.min,
            False,
            id="history-full-modified-eq-cursor-uri-lt"
        ),
        pytest.param(
            RemoteFile(uri="new.csv", last_modified=datetime.strptime("2020-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            True,
            (datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            datetime.min,
            True,
            id="history-full-modified-before-cursor-and-after-sync-start"
        ),
        pytest.param(
            RemoteFile(uri="new.csv", last_modified=datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT)),
            {},
            True,
            (datetime.strptime("2022-01-01T00:00:00.000000Z", DATE_TIME_FORMAT), "a.csv"),
            datetime.strptime("2024-01-01T00:00:00.000000Z", DATE_TIME_FORMAT),
            False,
            id="history-full-modified-before-cursor-and-before-sync-start"
        ),
    ]
)
def test_should_sync_file(
    file_to_check: RemoteFile,
    history: Dict[str, Any],
    is_history_full: bool,
    prev_cursor_value: Tuple[datetime, str],
    sync_start: datetime,
    expected_should_sync: bool,
):
    cursor = _make_cursor({})
    cursor._file_to_datetime_history = history
    cursor._prev_cursor_value = prev_cursor_value
    cursor._sync_start = sync_start
    cursor._is_history_full = MagicMock(return_value=is_history_full)
    should_sync = cursor._should_sync_file(file_to_check, MagicMock())
    assert should_sync == expected_should_sync


@freeze_time("2023-06-16T00:00:00Z")
@pytest.mark.parametrize(
    "input_history, is_history_full, expected_start_time",
    [
        pytest.param({}, False, datetime.min, id="empty-history"),
        pytest.param(
            {"a.csv": "2021-01-01T00:00:00.000000Z"},
            False,
            datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT),
            id="non-full-history"
        ),
        pytest.param(
            {f"file{i}.csv": f"2021-01-0{i}T00:00:00.000000Z" for i in range(1, 4)},  # all before the time window
            True,
            datetime.strptime("2021-01-01T00:00:00.000000Z", DATE_TIME_FORMAT),  # Time window start time
            id="full-history-earliest-before-window"
        ),
        pytest.param(
            {f"file{i}.csv": f"2024-01-0{i}T00:00:00.000000Z" for i in range(1, 4)},  # all after the time window
            True,
            datetime.strptime("2023-06-13T00:00:00.000000Z", DATE_TIME_FORMAT),  # Earliest file time
            id="full-history-earliest-after-window"
        ),
    ]
)
def test_compute_start_time(input_history, is_history_full, expected_start_time, monkeypatch):
    cursor = _make_cursor({"history": input_history})
    cursor._file_to_datetime_history = input_history
    cursor._is_history_full = MagicMock(return_value=is_history_full)
    assert cursor._compute_start_time() == expected_start_time
